#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Written by Paul Rougieux and Selene Patani.

Copyright (c) 2023 European Union
Licenced under the MIT licence

JRC biomass Project.
Unit D1 Bioeconomy.

Usage: download FAOSTAT datasets and store them in the database defined in
`faostat.db`:

    >>> from biotrade.faostat import faostat
    >>> faostat.pump.update(["crop_production", "crop_trade"])
    >>> faostat.pump.update(["forestry_production", "forestry_trade"])
    >>> faostat.pump.update(["food_balance"])
    >>> faostat.pump.update(["land_use", "land_cover"])
    >>> # Skip the table deletion confirmation message
    >>> faostat.pump.update(["land_use", "land_cover"], skip_confirmation=True)

"""

# Built-in modules
from pathlib import Path
from zipfile import ZipFile
import re
import shutil
import tempfile
import csv

try:
    import requests
except Exception as e:
    msg = "Failed to import requests, you will not be able to load data from FAOSTAT,"
    msg += "but you can still use other methods.\n"
    print(msg, str(e))

# Third party modules
import logging
import pandas
import numpy as np

# Internal modules
from biotrade.common.url_request_header import HEADER


def choose_column_renaming(short_name):
    """Choose which column from config_data/column_names.csv to use for
    renaming."""
    output = None
    for keyword in ["production", "trade", "land", "food_balance", "country"]:
        if keyword in short_name:
            output = "faostat_" + keyword
    if output is None:
        raise ValueError("No column to use for renaming", short_name)
    return output


class Pump:
    """
    Download trade data from FAOSTAT and store it locally in a database.

    The pump can perform the following tasks:
        1. Download compressed csv files from FAOSTAT.
        2. Read the compressed csv files into pandas data frames.
        3. Transfer the data frames to a database.

    An update performs all 3 tasks, it download files, reads them in chunks and
    stores them in the database

        >>> from biotrade.faostat import faostat
        >>> faostat.pump.update(["crop_production", "crop_trade"])
        >>> faostat.pump.update(["food_balance"])
        >>> faostat.pump.update(["land_use", "land_cover"])

    List available datasets and metadata links:

        >>> faostat.pump.datasets
        >>> faostat.pump.metadata_link

    Update all FAOSTAT datasets:

        >>> faostat.pump.download_all_datasets()
        >>> faostat.pump.transfer_all_datasets()
        >>> # Optionally, you can skip some of the large datasets
        >>> faostat.pump.transfer_all_datasets(skip=["food_balance", "crop_trade"])

    The following examples use lower level function which are not needed for
    normal use.

    Read an entire table directly from a CSV file to a data frame
    without going through the database:

        >>> fp = faostat.pump.read_df("forestry_production")

    Note that reading the entire dataset into a data frames can take a large
    part of the memory. It is recommended to start an analysis with a smaller
    data frame for a specific country or a specific product, using the
    faostat.db.select method.

    Update a dataset by downloading it again from FAOSTAT:

        >>> faostat.pump.download_zip_csv("Forestry_E_All_Data_(Normalized).zip")
        >>> fp = faostat.pump.read_df("forestry_production")
        >>> faostat.self.db.write_df(fp, "forestry_production")

    """

    # Log debug and error messages
    logger = logging.getLogger("biotrade.faostat")
    # Define URL request headers
    header = HEADER
    # Base URL to load data from the website
    url_api_base = "http://fenixservices.fao.org/faostat/static/bulkdownloads/"
    # Dataset names on the FAOSTAT platform https://www.fao.org/faostat/en/#data
    datasets = {
        "forestry_production": "Forestry_E_All_Data_(Normalized).zip",
        "forestry_trade": "Forestry_Trade_Flows_E_All_Data_(Normalized).zip",
        "forest_land": "Emissions_Land_Use_Forests_E_All_Data_(Normalized).zip",
        "crop_production": "Production_Crops_Livestock_E_All_Data_(Normalized).zip",
        "crop_trade": "Trade_DetailedTradeMatrix_E_All_Data_(Normalized).zip",
        "land_cover": "Environment_LandCover_E_All_Data_(Normalized).zip",
        "land_use": "Inputs_LandUse_E_All_Data_(Normalized).zip",
        "food_balance": "SUA_Crops_Livestock_E_All_Data_(Normalized).zip",
    }
    # Link to the metadata
    metadata_link = {
        "forestry_production": "https://www.fao.org/faostat/en/#data/FO/metadata",
        "forestry_trade": "https://www.fao.org/faostat/en/#data/FT/metadata",
        "forest_land": "https://www.fao.org/faostat/en/#data/GF/metadata",
        "crop_production": "https://www.fao.org/faostat/en/#data/QCL/metadata",
        "crop_trade": "https://www.fao.org/faostat/en/#data/TM/metadata",
        "land_cover": "https://www.fao.org/faostat/en/#data/LC/metadata",
        "land_use": "https://www.fao.org/faostat/en/#data/RL/metadata",
        "food_balance": "https://www.fao.org/faostat/en/#data/SCL/metadata",
    }

    def __init__(self, parent):
        # Default attributes #
        self.parent = parent
        self.db = self.parent.db
        self.data_dir = self.parent.data_dir
        # Mapping table used to rename columns
        self.column_names = self.parent.column_names
        # Number of lines to transfer from csv files to the database at once
        self.chunk_size = 10**7

    def download_zip_csv(self, zip_file_name):
        """Download a compressed csv file from the FAOSTAT website

         Example use:

         >>> from biotrade.faostat import faostat
         >>> faostat.pump.download_zip_csv("Forestry_E_All_Data_(Normalized).zip")
         >>> faostat.pump.download_zip_csv("Forestry_Trade_Flows_E_All_Data_(Normalized).zip")
         >>> faostat.pump.download_zip_csv("Production_Crops_Livestock_E_All_Data_(Normalized).zip")
         >>> faostat.pump.download_zip_csv("Trade_DetailedTradeMatrix_E_All_Data_(Normalized).zip")

        # Check the content of the destination folder for updates
        !ls -al ~/repos/biotrade_data/faostat/
        """
        url_api_call = self.url_api_base + zip_file_name
        output_file = self.data_dir / zip_file_name
        self.logger.info("Downloading data from:\n %s", url_api_call)
        response = requests.get(url=url_api_call, headers=self.header, stream=True)
        with open(output_file, "wb") as out_file:
            print(f"HTTP response code: {response.status_code}")
            shutil.copyfileobj(response.raw, out_file)

    def read_zip_csv_to_df(
        self, zip_file, column_renaming, short_name, encoding="latin1"
    ):
        """Read a zip file downloaded from the FAOSTAT API rename columns and return a data frame

        The zip file contains 2 csv file, a large one with the data and a small one with flags.
        We want to open the large csv file which has the same name as the zip file.

        Columns are renamed using the mapping table defined in config_data/column_names.csv.
        The product and element columns are renamed to snake case.

        Example use:

        >>> from biotrade.faostat import faostat
        >>> zip_file = faostat.data_dir / "Forestry_E_All_Data_(Normalized).zip"
        >>> df = faostat.pump.read_zip_csv_to_df(
        >>>     zip_file=zip_file,
        >>>     column_renaming="faostat_production",
        >>>     short_name = "forestry_production")

        """
        # Extract the name of the CSV file
        zip_file_name = Path(zip_file).name
        csv_file_name = re.sub(".zip$", ".csv", zip_file_name)
        # Read to a pandas data frame
        with ZipFile(zip_file) as zipfile:
            with zipfile.open(csv_file_name) as csvfile:
                df = pandas.read_csv(csvfile, encoding=encoding)
        df = self.sanitize_variable_names(df, column_renaming, short_name)
        return df

    def sanitize_variable_names(self, df, column_renaming, short_name):
        # TODO: use the function sanitize_variable_names common/sanitise.py
        """Sanitize column names using the mapping table.
        Use snake case in product and element names"""
        # Rename columns to snake case
        df.rename(columns=lambda x: re.sub(r"\W+", "_", str(x)).lower(), inplace=True)
        # Columns of the db table
        db_table_cols = self.db.tables[short_name].columns.keys()
        # Original column names
        cols_to_check = self.column_names[
            self.column_names["biotrade"].isin(db_table_cols)
        ][column_renaming].tolist()
        # Check columns which have changed in the input source
        cols_to_change = set(cols_to_check).difference(df.columns)
        # If column names have changed raise an error
        if cols_to_change:
            raise ValueError(
                f"The following columns \n{list(cols_to_change)}\nhave changed in the input source {column_renaming}.\nUpdate config_data/column_names.csv before updating table {short_name}"
            )
        # Map columns using the naming convention defined in self.column_names
        mapping = self.column_names.set_index(column_renaming).to_dict()["biotrade"]
        # Discard nan keys of mapping dictionary
        mapping.pop(np.nan, None)
        # Obtain columns for db upload
        columns = list(df.columns.intersection(list(mapping.keys())))
        # Filter df selecting only columns for db
        df = df[columns].copy()
        # Rename columns using the naming convention defined in self.column_names
        df.rename(columns=mapping, inplace=True)
        # Rename column contents to snake case using a compiled regex
        regex_pat = re.compile(r"\W+")
        for column in ["product", "item", "element"]:
            if column in df.columns:
                df[column] = (
                    df[column].str.replace(regex_pat, "_", regex=True).str.lower()
                )
                # Remove the last underscore if it's at the end of the name
                df[column] = df[column].str.replace("_$", "", regex=True)
        # Convert NaN flags to an empty character variable
        # so that the flag column doesn't get converted to a list column when sent to R
        # Here is how the flag was encoded before the change
        # ft.flag.unique()
        # array([None, '*', 'R', 'Cv', 'P', 'A'], dtype=object)
        # Because of the absence of na.character type in pandas
        # these appear as two different data types when sent to R and that column is then
        # converted to a list column
        # Here is how the flag is encoded after the change
        # ft.flag.unique()
        # array(['', '*', 'R', 'Cv', 'P', 'A'], dtype=object)
        if "flag" in df.columns:
            df.flag.fillna("", inplace=True)
        return df

    def read_df(self, short_name):
        """Read an entire zip csv file to a data frame

        Not recommended for large datasets which don't fit into memory.

        Example use:

        >>> from biotrade.faostat import faostat
        >>> fp = faostat.pump.read_df("forestry_production")
        >>> lu = faostat.pump.read_df("land_use")
        >>> lc = faostat.pump.read_df("land_cover")
        >>> fl = faostat.pump.read_df("forest_land")

        """
        # Read the compressed CSV into a data frame
        df = self.read_zip_csv_to_df(
            zip_file=self.data_dir / self.datasets[short_name],
            column_renaming=choose_column_renaming(short_name),
            short_name=short_name,
        )
        return df

    def transfer_csv_to_db_in_chunks(self, short_name, chunk_size):
        """Transfer large CSV files to the database in chunks
        so that a data frame with 40 million rows doesn't overload the memory.
        """
        temp_dir = Path(tempfile.TemporaryDirectory().name)
        # Csv file inside biotrade package config data directory
        if short_name == "country":
            csv_file_name = self.parent.config_data_dir / "faostat_country_groups.csv"
            encoding_var = "utf-8"
        # Zip files for table data
        else:
            # Unzip the CSV and write it to a temporary file on disk
            try:
                zip_file = ZipFile(self.data_dir / self.datasets[short_name])
                zip_file.extractall(temp_dir)
                csv_file_name = temp_dir / re.sub(
                    ".zip$", ".csv", self.datasets[short_name]
                )
                encoding_var = "latin1"
                # Test if the file is corrupted
                with open(csv_file_name, "r", encoding=encoding_var) as csvfile:
                    # Detect the delimiter
                    dialect = csv.Sniffer().sniff(csvfile.read(1024))
                    # Place the reader at the beginning
                    csvfile.seek(0)
                    # Read the file
                    reader = csv.reader(csvfile, dialect)
                    header = next(reader)
                    for row in reader:
                        pass
            # Zip file corrupted
            except Exception as e:
                self.db.logger.warning(
                    f"File for {short_name} table is not available due to {e}.\n Unable to pump {short_name} data."
                )
                if temp_dir.exists():
                    # Remove temporary directory
                    shutil.rmtree(temp_dir)
                return
        # Drop and recreate the table
        table = self.db.tables[short_name]
        table.drop(self.db.engine)
        self.db.create_if_not_existing(table)
        # Read in chunk and pass each chunk to the database
        for df_chunk in pandas.read_csv(
            csv_file_name, chunksize=chunk_size, encoding=encoding_var
        ):
            df_chunk = self.sanitize_variable_names(
                df_chunk, choose_column_renaming(short_name), short_name
            )
            print(df_chunk.head(1))
            self.db.append(df=df_chunk, table=short_name)
        if temp_dir.exists():
            # Remove temporary directory
            shutil.rmtree(temp_dir)

    def confirm_db_table_deletion(self, datasets):
        """Confirm database table deletion

        Separate method, because it is reused at different places."""
        msg = f"\nIf the database {self.db.engine} exists already, "
        msg += "this command will erase the following tables "
        msg += "and replace them with new data:\n - "
        msg += "\n - ".join(datasets)
        if input(msg + "\nPlease confirm [y/n]:") != "y":
            print("Cancelled.")
            return False
        else:
            return True

    def transfer_to_db(self, datasets, skip_confirmation=False):
        """Transfer from a csv file to the database by replacing the table
        content with the content of the zipped CSV files. Database field types
        are determined in faostat.db.

        :param list datasets: list of dataset names, whose keys should be in
            the faostat.pump.datasets and faostat.db.tables dictionaries
        :return: Nothing

        Usage:

            >>> from biotrade.faostat import faostat
            >>> faostat.pump.transfer_to_db()

        Use a larger chunk size

            >>> faostat.pump.chunk_size = 10 ** 6

        Skip the large crop trade and food balance datasets

            >>> faostat.pump.transfer_to_db(skip_crop_trade=True)

        """
        # Make datasets a list
        if isinstance(datasets, str):
            datasets = [datasets]
        if not skip_confirmation:
            if not self.confirm_db_table_deletion(datasets):
                return
        for table_name in datasets:
            # Transfer the compressed CSV file to the database
            self.transfer_csv_to_db_in_chunks(table_name, self.chunk_size)

    def update(self, datasets, skip_confirmation=False):
        """Update the given datasets by downloading them from FAOSTAT and
        transferring them to the database

        :param list or str datasets: list of dataset names, whose keys should
            be in the faostat.pump.datasets and faostat.db.tables dictionaries

        Usage:

            >>> from biotrade.faostat import faostat
            >>> faostat.pump.update(["crop_production", "crop_trade"])
            >>> faostat.pump.update(["food_balance"])
            >>> faostat.pump.update(["land_use", "land_cover"])

        """
        # Make datasets a list
        if isinstance(datasets, str):
            datasets = [datasets]
        # Confirmation message
        if not skip_confirmation:
            if not self.confirm_db_table_deletion(datasets):
                return
        # Download datasets from FAOSTAT
        for this_dataset in datasets:
            if this_dataset == "country":
                continue
            zip_file_name = self.datasets[this_dataset]
            self.download_zip_csv(zip_file_name)
        # Transfer to the database
        # Skip confirmation because we already confirmed above
        self.transfer_to_db(datasets, skip_confirmation=True)

    def download_all_datasets(self):
        """Download all files in the datasets dictionary"""
        for zip_file_name in self.datasets.values():
            self.download_zip_csv(zip_file_name)

    def transfer_all_datasets(self, skip):
        """Transfer all datasets to the database
        :param list skip: skip datasets in this list"""
        datasets = list(self.datasets.keys())
        # Make skip a list and remove list items from datasets
        if isinstance(skip, str):
            skip = [skip]
        for this_dataset in skip:
            datasets.remove(this_dataset)
        self.transfer_to_db(datasets)

    def show_metadata_link(self, short_name):
        """Display the metadata link associated with a dataset

        >>> from biotrade.faostat import faostat
        >>> faostat.pump.show_metadata_link("forestry_production")
        >>> faostat.pump.show_metadata_link("land_use")
        """
        print(self.metadata_link[short_name])



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Test data aggregation functions from:

    - common/aggregate.py
    - faostat/aggregate.py

Written by Paul Rougieux and Selene Patani.

Copyright (c) 2023 European Union
Licenced under the MIT licence
"""

import pandas
from pandas.testing import assert_frame_equal
from pandas.testing import assert_series_equal
from biotrade.faostat.aggregate import agg_trade_eu_row, agg_by_country_groups


def test_agg_by_country_groups():
    df = pandas.DataFrame(
        {
            "reporter": [
                "Spain",
                "Italy",
                "Afghanistan",
                "Australia",
                "Sweden",
                "Italy",
            ],
            "reporter_code": [203, 106, 2, 10, 210, 106],
            "partner": [
                "Argentina",
                "Colombia",
                "Brazil",
                "Norway",
                "Mexico",
                "Mexico",
            ],
            "partner_code": [9, 44, 21, 162, 138, 138],
            "value": [1, 2, 3, 5, 7, 1],
        }
    )
    dfp_expected = pandas.DataFrame(
        {
            "continent_reporter": ["Asia", "Europe", "Oceania"],
            "continent_partner": ["Americas", "Americas", "Europe"],
            "value": [3, 11, 5],
        }
    )
    dfp_output = agg_by_country_groups(
        df, agg_reporter="continent", agg_partner="continent"
    )
    assert_series_equal(dfp_output["value"], dfp_expected["value"])
    assert_frame_equal(dfp_output, dfp_expected)
    dfp_expected = pandas.DataFrame(
        {
            "reporter": [
                "Afghanistan",
                "Australia",
                "Italy",
                "Italy",
                "Spain",
                "Sweden",
            ],
            "reporter_code": [2, 10, 106, 106, 203, 210],
            "sub_continent_partner": [
                "South America",
                "Northern Europe",
                "Central America",
                "South America",
                "South America",
                "Central America",
            ],
            "value": [3, 5, 1, 2, 1, 7],
        }
    )
    dfp_output = agg_by_country_groups(df, agg_partner="sub_continent")
    assert_series_equal(dfp_output["value"], dfp_expected["value"])
    assert_frame_equal(dfp_output, dfp_expected)


def test_agg_trade_eu_row_with_letters():
    df = pandas.DataFrame(
        {
            "reporter": ["A", "A", "B"],
            "partner": ["Y", "Z", "Z"],
            "value": [1, 2, 1],
        }
    )
    dfp_expected = pandas.DataFrame(
        {"reporter": ["A", "B"], "partner": ["row", "row"], "value": [3, 1]}
    )
    dfp_output = agg_trade_eu_row(df, grouping_side="partner")
    assert_series_equal(dfp_output["value"], dfp_expected["value"])
    assert_frame_equal(dfp_output, dfp_expected)
    dfr_expected = pandas.DataFrame(
        {"reporter": ["row", "row"], "partner": ["Y", "Z"], "value": [1, 3]}
    )
    dfr_output = agg_trade_eu_row(df, grouping_side="reporter")
    assert_series_equal(dfr_output["value"], dfr_expected["value"])
    assert_frame_equal(dfr_output, dfr_expected)


def test_agg_trade_eu_row_with_eu_countries():
    df = pandas.DataFrame(
        {
            "reporter": ["Italy", "Italy", "Italy", "A"],
            "partner": ["Y", "France", "France", "Y"],
            "value": [1, 2, 1, 2],
        }
    )
    dfp_expected = pandas.DataFrame(
        {
            "reporter": ["A", "Italy", "Italy"],
            "partner": ["row", "eu", "row"],
            "value": [2, 3, 1],
        }
    )
    dfp_output = agg_trade_eu_row(df, grouping_side="partner")
    assert_series_equal(dfp_output["value"], dfp_expected["value"])
    assert_frame_equal(dfp_output, dfp_expected)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Written by Paul Rougieux and Selene Patani.

Copyright (c) 2023 European Union
Licenced under the MIT licence

JRC biomass Project.
Unit D1 Bioeconomy.

Parent object to faostat.database.py and comtrade.database.py

"""

from sqlalchemy import select
from sqlalchemy.sql import func
from sqlalchemy import UniqueConstraint


class Database:
    """
    Database parent class

    Keep methods and parameters commons to all databases in this file.
    """

    # To be overwritten by the children
    logger = None
    database_url = None
    engine = None
    inspector = None
    schema = None
    tables = None

    def check_unique_constraints(self, df, table):
        """ "
        Method that checks if nan values are present for unique constraint columns and delete the related rows to avoid duplications before appending the dataframe. In fact
        PostgreSQL does not consider nan values as unique and allows more than 1 insertion, causing duplications without errors.

        :param df, Dataframe to be uploaded
        :param table, string name of the table to upload the dataframe
        :return df, Dataframe without nan values in constraint columns

        """
        # Loop on the constraints
        for const in self.tables[table].constraints:
            # Check unique constraints
            if isinstance(const, UniqueConstraint):
                # Get the unique constraints columns
                unique_list = const.columns.keys()
                # Check if some of unique columns contains nan values
                df_nan = df[df[unique_list].isnull().any(axis=1)]
                # If nan values are present, raise a warning in the log and delete those rows
                if len(df_nan):
                    self.logger.warning(
                        "The following "
                        + str(len(df_nan))
                        + " rows will not be uploaded to the database table "
                        + table
                        + " due to nan values of unique constraints "
                        + str(unique_list)
                        + ":\nColumn names\n"
                        + str(df_nan.columns.tolist())
                        + "\nRows\n"
                        + "\n".join(str(i) for i in df_nan.values.tolist())
                    )
                    df.dropna(subset=unique_list, inplace=True)
                break
        return df

    def append(self, df, table, drop_description=True):
        """Store a data frame inside a given database table

        Note: we can only use df.to_sql() with the argument if_exists="append".
        Using if_exists="replace" would erase the table structure defined as
        SQlAlchemy metadata and use a default table structure based on data
        frame column types. We don't want the automated structure for several
        reasons. In particular, we want the database engine to enforce unique
        constraints and to return errors if data frame field types are not
        compatible with table field types defined in the database."""
        # Drop the lengthy product description
        if drop_description and "product_description" in df.columns:
            df.drop(columns=["product_description"], inplace=True)
        # Raise an error if the table is not defined in the metadata
        if table not in self.tables:
            raise ValueError(f"The table '{table}' is not defined in: \n{self}.")
        df = self.check_unique_constraints(df, table)
        df.to_sql(
            name=table,
            con=self.engine,
            schema=self.schema,
            if_exists="append",
            index=False,
            chunksize=10**6,
        )
        self.logger.info("Wrote %s rows to the database table %s", len(df), table)

    def create_if_not_existing(self, table):
        """Create a table in the database if it doesn't exist already

        table sqlalchemy.sql.schema.Table instance description of a table structure
        """
        #  Create the table if it doesn't exist
        if not self.inspector.has_table(table.name, schema=self.schema):
            table.create(bind=self.engine)
            self.logger.info("Created table %s in schema %s.", table.name, self.schema)
        return table

    def most_recent_year(
        self,
        table,
    ):
        """
        Query db table to check which is the most recent year of data.

        :param (string) table, name of table to check data

        :return (int) most_recent_year, the year of the most updated version of table data

        For example check which is the most recent year for yearly table of Comtrade db

            >>> from biotrade.comtrade import comtrade
            >>> most_recent_year = comtrade.db.most_recent_year("yearly")

        """
        # Select table
        table = self.tables[table]
        # Most recent year statement
        stmt = select(func.max(table.c.year))
        # Read data from db to data frame and select the first one
        most_recent_year = self.read_sql_query(stmt)
        most_recent_year = most_recent_year.values[0][0]
        return most_recent_year

    def confirm_db_table_deletion(self, datasets):
        """Confirm database table deletion

        Separate method, because it is reused at different places."""
        msg = f"\nIf the database {self.engine} exists already, "
        msg += "this command will erase the following tables "
        msg += "and replace them with new data:\n - "
        msg += "\n - ".join(datasets)
        if input(msg + "\nPlease confirm [y/n]:") != "y":
            print("Cancelled.")
            return False
        else:
            return True



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Written by Paul Rougieux and Selene Patani.

Copyright (c) 2023 European Union
Licenced under the MIT licence

JRC biomass Project.
Unit D1 Bioeconomy.

"""

# Third party modules
import pandas


class CountryGroups(object):
    """
    Comtrade product list, with additional information.

    Usage:

        >>> from biotrade.faostat import faostat
        >>> print("List of EU countries:")
        >>> print(faostat.country_groups.eu_country_names)
        >>> print("\n\nData frame of country groups:")
        >>> print(faostat.country_groups.df)

    Find the country codes for a given country using string matching:

        >>> country_codes = faostat.country_groups.search_name(["Brazil", "Indonesia"])
        >>> country_codes[["faost_code", "iso3_code", "un_code", "short_name"]]

    """

    def __init__(self, parent):
        # Default attributes #
        self.parent = parent
        # Directories #
        self.config_data_dir = self.parent.config_data_dir

    @property
    def df(self):
        """Country groupings

        :return: A data frame of table faostat_country_groups.csv

            >>> from biotrade.faostat import faostat
            >>> df = faostat.country_groups.df

        """
        path = self.config_data_dir / "faostat_country_groups.csv"
        df = pandas.read_csv(
            path,
            keep_default_na=False,
            na_values=[""],
        )
        # Force country codes to be integers instead of floats.
        # Note, we are not using the nullable integer data type Int64 with capital I.
        # https://pandas.pydata.org/pandas-docs/stable/user_guide/integer_na.html
        # Preferring to keep the standard integer type and -1 for missing codes.
        # TODO: remove this behaviour NA should be kept as NA
        df["faost_code"] = df["faost_code"].fillna(-1).astype("int")
        df["un_code"] = df["un_code"].fillna(-1).astype("int")
        return df

    @property
    def eu_country_names(self):
        """
        EU country name list in the FAOSTAT data
        :return list of eu country names

            >>> from biotrade.faostat import faostat
            >>> eu_country_name_list = faostat.country_groups.eu_country_names
        """
        df = self.df
        return df[df["eu27"] == 1]["fao_table_name"].tolist()

    @property
    def continents(self):
        """Country groupings by continents and subcontinents

        For example select from the crop trade table soy products and merge
        with the continents and subcontinents data, regarding both reporter
        and partner countries.

        >>> from biotrade.faostat import faostat
        >>> db = faostat.db
        >>> df_soy = db.select(table="crop_trade", product = "soy")
        >>> df_continents = faostat.country_groups.continents
        >>> df_soy_merge = df_soy.merge(df_continents, how='left',
        >>>     left_on = 'reporter_code', right_on = 'faost_code')
        >>> df_soy_merge = df_soy_merge.merge(df_continents, how='left',
        >>>     left_on = 'partner_code', right_on = 'faost_code',
        >>>     suffixes=('_reporter','_partner'))
        """
        return self.df[["faost_code", "continent", "sub_continent"]]

    def search_name(self, pattern: str) -> pandas.DataFrame:
        """Search for a country name in the country groups table"""
        if isinstance(pattern, str):
            pattern = [pattern]
        selector = self.df["short_name"].str.contains("|".join(pattern), case=False)
        return self.df.loc[selector.fillna(False)]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Written by Paul Rougieux and Selene Patani.

Copyright (c) 2023 European Union
Licenced under the MIT licence

JRC biomass Project.
Unit D1 Bioeconomy.

"""

# Internal modules
from biotrade.faostat import faostat

# Third party modules
from sqlalchemy.sql import func
from sqlalchemy import select
import pandas as pd


def compare_crop_production_harvested_area_world_agg_country(
    drop_identical=False, product_code=None
):
    """
    Function which compares Faostat crop production and harvested area with partner equal to World
    and country partner aggregations. It returns a dataframe containing the difference between aggregation values and World data reported by Faostat

    :param drop_identical (boolean), if True drop rows where the aggregation value match with the Word value reported by Faostat, otherwise it keeps all the rows
    """
    if isinstance(product_code, (int, str)):
        product_code = [product_code]
    # Crop production table to select from raw_faostat schema
    table = faostat.db.tables["crop_production"]
    # Columns to be considered before performing the join
    columns = [
        table.c.product_code,
        table.c.product,
        table.c.element_code,
        table.c.element,
        table.c.unit,
        table.c.period,
        table.c.year,
    ]
    # Add aggregation quantity column
    table_stmt = table.select().with_only_columns(
        [*columns, func.sum(table.c.value).label("qnt_agg")]
    )
    # Consider all reporters excluding China, regional, subcontinent and continent aggregations (reporter_code < 300)
    table_stmt = table_stmt.where(table.c.reporter_code < 300)
    # Consider specific product codes if defined
    if product_code is not None:
        table_stmt = table_stmt.where(table.c.product_code.in_(product_code))
    # Perform aggregation
    table_stmt = table_stmt.group_by(*columns)
    # Define the new table name
    agg_selection = table_stmt.subquery().alias("agg")
    # Perform the same operations for reporter_code = 5000 (World)
    table_stmt = table.select().with_only_columns(
        [
            *columns,
            func.sum(table.c.value).label("qnt_world"),
        ]
    )
    table_stmt = table_stmt.where(table.c.reporter_code == 5000)
    # Consider specific product codes if defined
    if product_code is not None:
        table_stmt = table_stmt.where(table.c.product_code.in_(product_code))
    table_stmt = table_stmt.group_by(*columns)
    world_selection = table_stmt.subquery().alias("world")
    # Full outer join to calculate the difference between results
    table_join = agg_selection.join(
        world_selection,
        (agg_selection.c.product_code == world_selection.c.product_code)
        & (agg_selection.c.element_code == world_selection.c.element_code)
        & (agg_selection.c.unit == world_selection.c.unit)
        & (agg_selection.c.period == world_selection.c.period),
        isouter=True,
        full=True,
    )
    table = select(table_join).alias("table_join")
    # Final columns to be retained
    join_columns = [
        table.c.product_code,
        table.c.product,
        table.c.element_code,
        table.c.element,
        table.c.unit,
        table.c.period,
        table.c.year,
        table.c.qnt_agg,
        table.c.qnt_world,
        (table.c.qnt_agg - table.c.qnt_world).label("qnt_diff"),
    ]
    table_stmt = select(join_columns)
    # Return all crop production and harvested area data
    table_stmt = table_stmt.where(table.c.unit.in_(["Head", "1000 No", "tonnes", "ha"]))
    if drop_identical:
        # Return crop production and harvested area data where aggregation quantities don't match with World data provided by Faostat
        # Null values are not counted in comparison with standard sql, added a separately statement
        table_stmt = table_stmt.where(
            (table.c.qnt_agg != table.c.qnt_world)
            | ((table.c.qnt_agg is None) & (table.c.qnt_world is not None))
            | ((table.c.qnt_agg is not None) & (table.c.qnt_world is None))
        )
    # Return the dataframe from the query to db
    df = pd.read_sql_query(table_stmt, faostat.db.engine)
    return df


def compare_prod_agg_to_parts(df, prod_aggregate, prod_constituents):
    """Compare aggregate products to the sum of their constituents

    :param DataFrame df: FAOSTAT production data in long format
    :param str prod_aggregate: Name of the aggregate product
    :param list prod_constituents: Names of the product constituents

    Return a reshaped data frame in wider format with aggregates and their constituents in columns.

    For more information, see the FAOSTAT forest products classification and
    standards at https://www.fao.org/forestry/statistics/80572/en/
    in particular the data structure chart
    https://www.fao.org/forestry/49962-0f43c0da7039a611aa884b3c6c642f4ac.pdf
    explains how the different products and aggregates relate to each other.

    For example compare roundwood to the sum of industrial roundwood and fuel
    wood. In the complete forestry production dataset (including production,
    import and export values and quantities).

    >>> from biotrade.faostat import faostat
    >>> from biotrade.faostat.quality import compare_prod_agg_to_parts
    >>> fp = faostat.pump.read_df("forestry_production")
    >>> rwd = compare_prod_agg_to_parts(
    >>>     fp, "roundwood", ["industrial_roundwood", "wood_fuel"]
    >>> )
    >>> assert all(rwd["diff"] == 0)

    """
    prod_together = [prod_aggregate] + prod_constituents
    index = ["reporter", "element", "unit", "year"]
    df = (
        df.query("product in @prod_together")
        .pivot(index=index, columns="product", values="value")
        .fillna(0)
    )
    df[prod_aggregate + "_agg"] = df.drop(columns=prod_aggregate).sum(axis=1)
    df["diff"] = df[prod_aggregate] - df[prod_aggregate + "_agg"]
    return df


def check_prod_agg_and_parts(df, prod_aggregate, prod_constituents):
    """Check that the product aggregate corresponds to the sum of its constituents

    This function has the same arguments as compare_prod_agg_to_parts. In case
    of an error, it gives the function call necessary to show the comparison
    data frame for further investigation.
    """
    output_df = compare_prod_agg_to_parts(
        df, prod_aggregate=prod_aggregate, prod_constituents=prod_constituents
    )
    if all(output_df["diff"] == 0):
        msg = f"OK: the product aggregate {prod_aggregate} "
        msg += f"does correspond to the sum of its constituents:\n {prod_constituents}"
        print(msg)
    else:
        msg = f"The product aggregate {prod_aggregate} "
        msg += (
            f"does not correspond to the sum of its constituents:\n {prod_constituents}"
        )
        msg += "\nLoad the data frame with:\n"
        msg += "from biotrade.faostat.quality import compare_prod_agg_to_parts\n"
        msg += f"compare_prod_agg_to_parts(your_dataframe, '{prod_aggregate}', {prod_constituents})"
        raise ValueError(msg)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Written by Paul Rougieux and Selene Patani.

Copyright (c) 2023 European Union
Licenced under the MIT licence

JRC biomass Project.
Unit D1 Bioeconomy.

Get data frames containing both FAOSTAT and Comtrade data for sawnwood and
soybean related products. In these data frames, the Comtrade product and
country codes have been converted to their equivalent FAOSTAT codes.

    >>> from biotrade.common.compare import merge_faostat_comtrade
    >>> swd = merge_faostat_comtrade(faostat_table="forestry_trade",
    >>>                              comtrade_table="yearly",
    >>>                              faostat_code = [1632, 1633])
    >>> soy = merge_faostat_comtrade(faostat_table="crop_trade",
    >>>                              comtrade_table="yearly",
    >>>                              faostat_code = [236, 237, 238])

Select the same data from Comtrade and compare to check that the
merge_faostat_comtrade() method doesn't modify the Comtrade data.

    >>> from biotrade.common.products import comtrade_faostat_mapping
    >>> from biotrade.comtrade import comtrade
    >>> import numpy as np
    >>> faostat_product = [236, 237, 238]
    >>> code_map = comtrade_faostat_mapping.query("faostat_code.isin(@faostat_product)")
    >>> soy_comtrade = (
    >>>     comtrade.db.select("yearly", product_code = code_map["comtrade_code"])
    >>>     .merge(comtrade_faostat_mapping.rename(columns={"comtrade_code":"product_code"}),
    >>>            on="product_code")
    >>> )
    >>> index = ["year"]
    >>> index = ["year", "reporter", "partner"]
    >>> # TODO: make it work also for country codes, requires a merge with FAOSTAT codes
    >>> # index = ["year", "reporter_code", "partner_code"]
    >>> soy_agg = (soy
    >>>     .query("source =='comtrade' and element.str.contains('quantity')")
    >>>     .groupby(["product_code", "element"] + index)["value"].sum()
    >>>     .reset_index()
    >>>     .assign(flow = lambda x: x["element"].str.replace("_quantity",""))
    >>>     .rename(columns={"product_code":"faostat_code"})
    >>> )
    >>> soy_comtrade_agg = (
    >>>     soy_comtrade.groupby(["faostat_code", "flow"] + index)["net_weight"].sum()
    >>>     .reset_index()
    >>>     .merge(soy_agg, on=["faostat_code", "flow"] + index, how="outer", indicator=True)
    >>> )
    >>> np.testing.assert_allclose(soy_comtrade_agg.query("_merge=='both'")["net_weight"],
    >>>                            soy_comtrade_agg.query("_merge=='both'")["value"])
    >>> # Reporter and partner which are in the Comtrade data frame but not in
    >>> # the data returned by merge_faostat_comtrade()
    >>> soy_comtrade_agg.query("_merge=='left_only'")
    >>> soy_comtrade_agg.query("_merge=='right_only'")

Use monthly Comtrade data:

    >>> swd = merge_faostat_comtrade(faostat_table="forestry_trade",
    >>>                              comtrade_table="monthly",monthly
    >>>                              faostat_code = [1632, 1633])

"""

import warnings
import pandas
import numpy as np
import pandas as pd
import math
from biotrade.faostat import faostat
from biotrade.comtrade import comtrade
from biotrade.common.products import comtrade_faostat_mapping

ELEMENT_DICT = {
    "element_code": [5610, 5622, 5910, 5922],
    "element": ["import_quantity", "import_value", "export_quantity", "export_value"],
}


def replace_exclusively(df, code_column, code_dict, na_value=-1):
    """Replace codes with the dict and all other values with na_value

    :param series pandas series of product or country codes to replace
    :param dict dict of key value replacement pairs sent to the
        pandas.core.series.replace method.
    :return a panda series with replaced values

    Not available values (na_value) are represented as -1 by default, because it is
    easier to handle as grouping variable.

    """
    selector = df[code_column].isin(code_dict.keys())
    # Extract the corresponding names to get a nicer warning when available
    name_column = code_column.replace("_code", "")
    if name_column in df.columns:
        missing = df.loc[~selector, [code_column, name_column]].drop_duplicates()
    else:
        missing = df.loc[~selector, [code_column]].drop_duplicates().drop_duplicates()
    # If there are missing values display a warning and set the code to na_value
    if not missing.empty:
        warnings.warn(
            f"The following codes are present in {code_column} but missing "
            f"from the mapping dictionary:\n{missing}"
        )
        # Add missing keys to the dictionary and map them to the na_value
        code_dict = code_dict.copy()
        code_dict.update(dict(zip(missing[code_column], [na_value] * len(missing))))
    return df[code_column].replace(code_dict)


def transform_comtrade_using_faostat_codes(
    comtrade_table, faostat_code=None, comtrade_code=None, aggregate=True
):
    """Load and transform a Comtrade data table to use FAOSTAT product, country codes and names

    :param comtrade_table str: name of the comtrade table to select from
    :param list faostat_code: list of faostat codes to be loaded, default is None
    :param list comtrade_code: list of comtrade codes to be loaded, default is None
    :param boolean aggregate: data are aggregagted or not by product code, default is True
    into faostat codes, default is True

    The function makes Comtrade monthly data available with FAOSTAT codes. It
    also works on Comtrade yearly data.
    It does the following:

        1. Find the corresponding Comtrade codes using the mapping table
        2. Load Comtrade monthly data for the corresponding codes
        3. Replace comtrade country names by the FAOSTAT names
        4. Replace product codes and country codes by the FAOSTAT codes
        5. Reshape the Comtrade data to a longer format similar to FAOSTAT
        6. Aggregate Comtrade flows to the FAOSTAT product codes

    Example use:

        >>> from biotrade.common.compare import transform_comtrade_using_faostat_codes

    """
    # 1. Find the corresponding Comtrade codes using the mapping table
    if faostat_code:
        # Define the mapping codes of Faostat
        if comtrade_code is None:
            selector = comtrade_faostat_mapping["faostat_code"].isin(faostat_code)
            product_mapping = comtrade_faostat_mapping[selector]
            comtrade_code = product_mapping["comtrade_code"]
    elif comtrade_code is None:
        raise ValueError("You need to specify at least a faostat or comtrade code")
    # 2. Load Comtrade data for the corresponding codes
    df_wide = comtrade.db.select(comtrade_table, product_code=comtrade_code)
    # Replace Comtrade country codes by the FAOSTAT country codes
    country_mapping = faostat.country_groups.df[["faost_code", "un_code"]]
    country_dict = country_mapping.set_index("un_code").to_dict()["faost_code"]
    df_wide["reporter_code"] = replace_exclusively(
        df_wide, "reporter_code", country_dict
    )
    df_wide["partner_code"] = replace_exclusively(df_wide, "partner_code", country_dict)
    # Monthly and Yearly Comtrade data may slightly differ for country names. Use Faostat country names for consistency
    country_table = faostat.db.select("country")
    df_wide = df_wide.merge(
        country_table[["country_code", "country_name"]],
        how="left",
        left_on="reporter_code",
        right_on="country_code",
    )
    df_wide["reporter"] = df_wide["country_name"]
    df_wide.drop(columns=["country_code", "country_name"], inplace=True)
    df_wide = df_wide.merge(
        country_table[["country_code", "country_name"]],
        how="left",
        left_on="partner_code",
        right_on="country_code",
    )
    df_wide["partner"] = df_wide["country_name"]
    df_wide.drop(columns=["country_code", "country_name"], inplace=True)
    # Reshape Comtrade to long format
    index = [
        "reporter_code",
        "reporter",
        "partner_code",
        "partner",
        "product_code",
        "period",
        "year",
        "unit",
        "flag",
        "flow_code",
        "flow",
    ]
    df = df_wide.melt(
        id_vars=index,
        # We loose the quantity column, but it is not available
        # in the monthly data see comtrade/database.py
        value_vars=["net_weight", "trade_value"],
        var_name="element",
        value_name="value",
    )
    # Add units
    df["unit"] = df["element"].replace({"net_weight": "kg", "trade_value": "usd"})
    # A query of the Comtrade monthly data shows that the "quantity" column is always empty
    #     select * from raw_comtrade.monthly where quantity is not null limit 4;
    # Returns 0 rows
    # Rename element="net_weight" to "quantity" to be similar to FAOSTAT
    df["element"] = (
        df["flow"].str.lower().replace("-", "_", regex=True) + "_" + df["element"]
    )
    df["element"] = (
        df["element"]
        .replace("_trade", "", regex=True)
        .replace("_net_weight", "_quantity", regex=True)
    )
    # Aggregate
    if aggregate:
        # Replace Comtrade product codes by the FAOSTAT product codes
        product_dict = product_mapping.set_index("comtrade_code").to_dict()[
            "faostat_code"
        ]
        df["product_code"] = replace_exclusively(df, "product_code", product_dict)
        # The "flag" column is kept out of the index so lines with different flags can be aggregated
        index = [
            "reporter_code",
            "reporter",
            "partner_code",
            "partner",
            "product_code",
            "period",
            "year",
            "unit",
            "element",
        ]
        df_agg = df.groupby(index, dropna=False)["value"].agg("sum").reset_index()
        # Check that the Comtrade data didn't change after aggregation
        assert math.isclose(df.value.sum(), df_agg.value.sum())
        df = df_agg
    return df


def merge_faostat_comtrade(
    faostat_table,
    comtrade_table,
    faostat_code=None,
    comtrade_code=None,
    aggregate=True,
    strict=True,
):
    """Merge faostat and comtrade bilateral trade data.

    :param faostat_table str: name of the faostat table to select from
    :param comtrade_table str: name of the comtrade table to select from
    :param list faostat_code: list of faostat codes to be loaded
    :param list comtrade_code: list of comtrade codes to be loaded, default is None and in this case they are mapped into faostat codes
    :param boolean aggregate: data are aggregated or not by product code, default is True
    :param boolean strict: whether to raise an error or a warning on duplicated country

    The function does the following:

        1. Load FAOSTAT bilateral trade data for the given codes
        2. Load the transformed version of the Comtrade data with faostat codes
            using the method `transform_comtrade_using_faostat_codes`
        3. Aggregate Comtrade from monthly to yearly. For the last data point
            extrapolate to the current year based on values from the last 12 months
        4. Concatenate FAOSTAT and Comtrade data

    For example load palm oil data for both Faostat and Comtrade (yearly):

        >>> from biotrade.common.compare import merge_faostat_comtrade
        >>> palm_comp = merge_faostat_comtrade(faostat_table="crop_trade",
        >>>                                    comtrade_table="yearly",
        >>>                                    faostat_code = [257])

    For example load sawnwood data from both Faostat and Comtrade (monthly):

        >>> from biotrade.common.compare import merge_faostat_comtrade
        >>> swd = merge_faostat_comtrade(faostat_table="forestry_trade",
        >>>                              comtrade_table="monthly",
        >>>                              faostat_code = [1632, 1633])

    To investigate the number of periods reported for each country in the most
    recent years, I used:

        >>> df_comtrade = transform_comtrade_using_faostat_codes(
        >>>     comtrade_table="monthly", faostat_code = [1632, 1633])
        >>> (df_comtrade.query("year >= year.max() -2")
        >>>  .groupby(["reporter", "period"])["value"].agg("sum")
        >>>  .reset_index()
        >>>  .value_counts(["reporter"])
        >>>  .reset_index().to_csv("/tmp/value_counts.csv")
        >>> )

    Max reporting period:

        >>> (df_comtrade.groupby("reporter")["period"].max()
        >>>  .to_csv("/tmp/max_period.csv"))

    """
    # 1. Load FAOSTAT bilateral trade data for the given codes
    if faostat_code:
        df_faostat = faostat.db.select(faostat_table, product_code=faostat_code)
        product_names = df_faostat[["product_code", "product"]].drop_duplicates()
        element_df = pd.DataFrame.from_dict(ELEMENT_DICT)
        # Convert trade values from 1000 USD to USD
        codes = element_df[element_df["element"].str.endswith("_value")][
            "element_code"
        ].values.tolist()
        selector = df_faostat["element_code"].isin(codes)
        df_faostat.loc[selector, "value"] = df_faostat.loc[selector, "value"] * 1e3
        df_faostat.loc[selector, "unit"] = "usd"
        # Convert tonnes to kg
        codes = element_df[element_df["element"].str.endswith("_quantity")][
            "element_code"
        ].values.tolist()
        selector = df_faostat["element_code"].isin(codes)
        df_faostat.loc[selector, "value"] = df_faostat.loc[selector, "value"] * 1e3
        df_faostat.loc[selector, "unit"] = "kg"
    else:
        df_faostat = pandas.DataFrame(
            columns=[
                "reporter_code",
                "reporter",
                "partner_code",
                "partner",
                "product_code",
                "product",
                "element_code",
                "element",
                "period",
                "year",
                "unit",
                "value",
                "flag",
            ]
        )
    # 2. Load Comtrade bilateral trade data for the given codes
    df_comtrade = transform_comtrade_using_faostat_codes(
        comtrade_table=comtrade_table,
        faostat_code=faostat_code,
        comtrade_code=comtrade_code,
        aggregate=aggregate,
    )
    if comtrade_table == "monthly":
        # 3. Aggregate Comtrade from monthly to yearly. For the last data point
        #    extrapolate to the current year based on values from the last 12 months
        # Group by year and compute the sum of values for the 12 month in each year
        index = [
            "reporter_code",
            "reporter",
            "partner_code",
            "partner",
            "product_code",
            "year",
            "unit",
            "element",
        ]
        df_comtrade_agg = df_comtrade.groupby(index)["value"].agg("sum")
        # The last year is not necessarily complete and it might differ by
        # countries. For any country. Sum the values of the last 12 months instead.
        # We need to go back a bit further , because in March of 2022, there might
        # be advanced countries which reported January 2022, but other countries
        # which still have their last reporting period as June 2021, or even
        # further back in 2020.
        df_comtrade = df_comtrade.copy()  # .query("year >= year.max() - 3").copy()
        df_comtrade["max_period"] = df_comtrade.groupby("reporter")["period"].transform(
            max
        )
        df_comtrade["last_month"] = df_comtrade["max_period"] % 100
        df_comtrade["previous_year"] = df_comtrade["max_period"] // 100 - 1
        # For the special case of December, last year stays the same
        # last month is zero so that 0+1 becomes January
        is_december = df_comtrade["last_month"] == 12
        df_comtrade.loc[is_december, "previous_year"] = df_comtrade["max_period"] // 100
        df_comtrade.loc[is_december, "last_month"] = 0
        df_comtrade["max_minus_12"] = (
            df_comtrade["previous_year"] * 100 + df_comtrade["last_month"] + 1
        )
        df_recent = df_comtrade.query("period >= max_minus_12").copy()
        df_recent["year"] = df_recent["previous_year"] + 1
        df_recent_agg = df_recent.groupby(index)["value"].agg(value_est=sum)
        # Combine the aggregated yearly values with the estimate for the last year
        df = pandas.concat([df_comtrade_agg, df_recent_agg], axis=1).reset_index()
    else:
        # Add column value estimation with nan values since yearly data are not estimated and drop column period
        df = df_comtrade
        df["value_est"] = np.nan
        df.drop(columns="period", inplace=True)
    # Replace value by the estimate "value_est" where it is defined
    selector = ~df.value_est.isna()
    df.loc[selector, "value"] = df.loc[selector, "value_est"]
    df.drop(columns="value_est", inplace=True)
    # Add the column flag for the estimates
    df["flag"] = np.nan
    df.loc[selector, "flag"] = "estimate"
    df.loc[~selector, "flag"] = ""
    if comtrade_code is None:
        # Add FAOSTAT product names
        df = df.merge(product_names, on="product_code")
    # Add faostat element codes to comtrade data for consistency
    cols = ["element_code", "element"]
    element_code_faostat = pandas.DataFrame(columns=cols)
    db = faostat.db
    # Retrieve element codes of Faostat db
    for table_name in ["crop_trade", "forestry_trade"]:
        table = db.tables[table_name]
        element_code = db.read_sql_query(
            table.select()
            .distinct(table.c.element_code, table.c.element)
            .with_only_columns(table.c.element_code, table.c.element)
        )
        element_code_faostat = pandas.concat(
            [element_code_faostat, element_code], ignore_index=True
        )
    element_code_faostat = element_code_faostat.drop_duplicates().reset_index(drop=True)
    element_code_faostat = (
        element_code_faostat[
            element_code_faostat["element_code"].isin(ELEMENT_DICT["element_code"])
        ].sort_values(by="element_code")
    ).reset_index(drop=True)
    # Put as int element code
    element_code_faostat["element_code"] = element_code_faostat["element_code"].astype(
        "int64"
    )
    # Put a check on element codes
    if not element_code_faostat.equals(pd.DataFrame.from_dict(ELEMENT_DICT)):
        warnings.warn(
            f"Element types for trade have changed from\n{pd.DataFrame.from_dict(ELEMENT_DICT)}\nPlease check them"
        )
    # Merge on element
    df = df.merge(element_code_faostat, how="left", on=["element"])
    # Fill nan with -1 to let potential joins on element_code column too for comtrade data
    df["element_code"] = df["element_code"].fillna(-1).astype(int)
    # 4. Concatenate FAOSTAT and Comtrade data
    df_faostat.drop(columns="period", inplace=True)
    df_faostat["source"] = "faostat"
    df["source"] = "comtrade"
    df_concat = pandas.concat([df_faostat, df])
    df_concat = df_concat.reset_index(drop=True)
    # Place the "source" column first for readability
    cols = df_concat.columns.to_list()
    cols = [cols[-1]] + cols[:-1]
    df_concat = df_concat[cols]
    # Raise a warning if there are duplicates of country names associated to the same country code
    for col in ["reporter", "partner"]:
        country_code_unique = df_concat.drop_duplicates(subset=[col, col + "_code"])[
            [col, col + "_code"]
        ]
        duplicates = country_code_unique[
            country_code_unique.duplicated(subset=col + "_code", keep=False)
        ]
        if len(duplicates):
            msg = (
                "There is more than 1 country code match for a country name"
                + f"\n{duplicates.sort_values(col+'_code').values}\n"
                + "These country names are present in the FAOSTAT data:\n"
                + f"{set(duplicates[col]) & set(df_faostat[col].unique())}.\n"
                + "These country names are present in the Comtrade table "
                + "(using replacement names from the FAOSTAT country table):\n"
                + f"{set(duplicates[col]) & set(df[col].unique())}\n\n"
                + "To fix this issue:\n"
                + "1) Update the FAOSTAT country table with:\n"
                + "    >>> from biotrade.faostat import faostat\n"
                + "    >>> faostat.pump.update('country')\n"
                + "2) Update the FAOSTAT trade data with:\n"
                + "    >>> from biotrade.faostat import faostat\n"
                + f"    >>> faostat.pump.update(['{faostat_table}'])\n"
                + "3) If the issue persist after performing the above two points,\n"
                + "   The following file needs to be changed:\n"
                + "   biotrade/biotrade/config_data/faostat_country_groups.csv\n"
                + "   followed by\n"
                + "    >>> faostat.pump.update('country')\n"
            )
            if strict:
                msg += "\nUse the argument strict=False to ignore this error."
                raise ValueError(msg)
            if not strict:
                warnings.warn(msg)
    return df_concat



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Written by Paul Rougieux and Selene Patani

Copyright (c) 2023 European Union
Licenced under the MIT licence

JRC biomass Project.
Unit D1 Bioeconomy.

"""

import numpy as np

# Internal modules
from biotrade.faostat import faostat

# Make agg_trade_eu_row available here for backward compatibility
# so that the following import statement continues to work:
# >>> from biotrade.faostat.aggregate import agg_trade_eu_row
from biotrade.common.aggregate import (  # noqa # pylint: disable=unused-import
    agg_trade_eu_row,
)

# Import country table selecting continents and sub continents columns
CONTINENTS = faostat.country_groups.continents[
    ["faost_code", "continent", "sub_continent"]
]


def agg_by_country_groups(df, agg_reporter=None, agg_partner=None):
    """
    Aggregate country data to continent or subcontinent groups

    :param data frame df: data from faostat merged with continent/subcontinent
    table
    :param str agg_reporter: "continent" or "sub_continent" defines which level
    of data aggregation to perform on the reporter side
    :param str agg_partner: "continent" or "sub_continent" defines which level
    of data aggregation to perform on the partner side
    :return dataframe aggregated by continent/subcontinent instead of
    countries

    For example selecting soy trade data for all world countries, aggregate
    data by continents or subcontinents for both reporter and partner

    Import function agg_by_country_groups

        >>> from biotrade.faostat import faostat
        >>> from biotrade.faostat.aggregate import agg_by_country_groups

    Select bilateral trade of soy

        >>> db = faostat.db
        >>> soy = db.select(table="crop_trade", product = "soy")

    Aggregate data by continents

        >>> soy_agg_continent = agg_by_country_groups(soy,
            agg_reporter = 'continent', agg_partner = 'continent')

    Aggregate data by subcontinents

        >>> soy_agg_subcontinent = agg_by_country_groups(soy,
            agg_reporter = 'sub_continent', agg_partner = 'sub_continent')

    For example selecting Brazil and Indonesia as reporter countries.
    Aggregate data by reporter countries and continents for partners only

        >>> soy = db.select(table="crop_trade",
                reporter = ["Brazil", "Indonesia"], product = "soy")
        >>> soy_agg_continent_r_p = agg_by_country_groups(soy,
                agg_partner = "continent")

    """

    # Consider countries with reporter code < 300, above to this value reporters are aggregations
    df = df[df.reporter_code < 300]
    # Merge reporters with the corresponding continent/subcontinent data
    df = df.merge(
        CONTINENTS,
        how="left",
        left_on="reporter_code",
        right_on="faost_code",
    )

    # Merge partners with the corresponding continent/subcontinent data
    if "partner_code" in df.columns:
        # Consider countries with partner code < 300, above to this value partners are aggregations
        df = df[df.partner_code < 300]
        df = df.merge(
            CONTINENTS,
            how="left",
            left_on="partner_code",
            right_on="faost_code",
            suffixes=("_reporter", "_partner"),
        )

    # fixed aggregation column names
    columns = [
        "period",
        "product",
        "product_code",
        "element",
        "element_code",
        "unit",
    ]
    # Columns for the aggregation
    index = []
    for col in columns:
        if col in df.columns:
            index.append(col)

    if agg_reporter is None:
        # aggregate by reporter and reporter code (if columns exist) since no
        # continent/subcontinent reporter aggregation selected
        [
            index.append(column)
            for column in df.columns
            if column in ["reporter", "reporter_code"]
        ]
    else:
        # check if the are reporter continent/subcontinent columns into
        # dataframe to aggregate
        [
            index.append(column)
            for column in df.columns
            if column in [f"{agg_reporter}_reporter", f"{agg_reporter}"]
        ]
    if agg_partner is None:
        # aggregate by partner and partner code (if columns exist) since no
        # continent/subcontinet partner aggregation selected
        [
            index.append(column)
            for column in df.columns
            if column in ["partner", "partner_code"]
        ]
    else:
        # check if the are partner continent/subcontinent columns into
        # dataframe to aggregate
        [
            index.append(column)
            for column in df.columns
            if column == f"{agg_partner}_partner"
        ]
    # Aggregate
    df_agg = df.groupby(index, dropna=False).agg(value=("value", "sum")).reset_index()
    # Check that the total value isn't changed
    np.testing.assert_allclose(
        df_agg["value"].sum(),
        df["value"].sum(),
        err_msg=f"The total value sum of the aggregated data {df_agg['value'].sum()}"
        + f" doesn't match with the sum of the input data frame {df['value'].sum()}",
    )
    return df_agg



"""
Written by Selene Patani.

Copyright (c) 2023 European Union
Licenced under the MIT licence

Script made to export data related to harvested area/production of countries associated to the regulation and commodity tree products, for the web platform

"""


def main():
    from scripts.front_end.functions import (
        main_product_list,
        filter_production_data,
        aggregated_data,
        reporter_iso_codes,
        replace_zero_with_nan_values,
        save_file,
    )
    from biotrade.faostat import faostat
    import pandas as pd

    # Obtain the main product codes
    main_product_list = main_product_list(["crop_production", "forestry_production"])
    # Select quantities from Faostat db for crop data for all countries (code < 1000)
    crop_data = faostat.db.select(
        table="crop_production",
        product_code=main_product_list,
        element=["production", "area_harvested", "stocks", "yield"],
    )
    # Select wood production data
    wood_data = faostat.db.select(
        table="forestry_production",
        product_code=main_product_list,
        element=["production"],
    )
    # Merge data
    crop_data = pd.concat([crop_data, wood_data], ignore_index=True)
    # Filter at country level and change units
    crop_data = filter_production_data(crop_data)
    # Aggregate french territories values to France and add them to the dataframe
    code_list = [68, 69, 87, 135, 182, 270, 281]
    agg_country_code = 68
    agg_country_name = faostat.country_groups.df
    agg_country_name = agg_country_name[
        agg_country_name.faost_code == agg_country_code
    ].fao_table_name.values[0]
    crop_data = aggregated_data(
        crop_data, code_list, agg_country_code, agg_country_name
    )
    # Substitute faostat codes with iso3 codes
    crop_data = reporter_iso_codes(crop_data)
    # Columns to be retained
    column_list = ["reporter_code", "product_code", "period", "value", "unit"]
    # Harvested area data
    dropna_col = ["value"]
    crop_data = replace_zero_with_nan_values(crop_data, dropna_col)
    crop_data = crop_data.dropna(subset=dropna_col)
    harvested_area = crop_data[crop_data["element"] == "area_harvested"][column_list]
    # Production data
    production = crop_data[crop_data["element"].isin(["production", "stocks"])][
        column_list
    ]
    # Yield data
    yields = crop_data[crop_data["element"] == "yield"][column_list]
    # Save csv files to env variable path or into biotrade data folder
    save_file(harvested_area, "harvested_area_annual_variation.csv")
    save_file(production, "production_annual_variation.csv")
    save_file(yields, "yield_annual_variation.csv")


# Needed to avoid running module when imported
if __name__ == "__main__":
    main()



"""
Written by Selene Patani.

Copyright (c) 2023 European Union
Licenced under the MIT licence

Functions to return time series change detections and associated plots.
"""

import math
import pandas as pd
import numpy as np
from scipy import optimize, stats
import matplotlib.pyplot as plt
import pymannkendall as mk
from collections import OrderedDict
from functools import partial
from multiprocessing import Pool


def plot_relative_absolute_change(df):
    """
    Function to plot relative and absolute change values (see relative_absolute_change function for the calculations)
    for each group of data
    :param (DataFrame) df, the output of relative_absolute_change function

    For example, select production quantity of palm oil from 1986 with Malaysia as reporter, compute the change values with respect
    to the previous five years and plot results

        >>> from biotrade.faostat import faostat
        >>> from biotrade.common.time_series import relative_absolute_change, plot_relative_absolute_change
        >>> palm_oil_data = faostat.db.select(
        >>>     table="crop_production", reporter="Malaysia", product_code=257
        >>> )
        >>> palm_oil_prod = palm_oil_data[
        >>>     (palm_oil_data["element"] == "production") & (palm_oil_data["year"] > 1986)
        >>> ]
        >>> df_change = relative_absolute_change(palm_oil_prod, last_value=False)
        >>> plot_relative_absolute_change(df_change)

    """
    # Group of columns to be considered
    group_data_pivot_table = [
        "reporter",
        "unit",
        "product_code",
    ]
    # Depending on the df, other columns to be considered in order to group the data
    add_group_data_column = [
        "product",
        "partner",
        "element",
        "flow",
        "unit_code",
    ]
    # Select for grouping the data only columns contained into the df
    for column in df:
        if column in add_group_data_column:
            # If Comtrade db, unit code and unit are nan and have to be removed
            if column == "unit_code":
                group_data_pivot_table.remove("unit")
                continue
            group_data_pivot_table.append(column)
    # Reshape df to plot
    df_reshape = df.pivot_table(
        index="year",
        columns=group_data_pivot_table,
        values=["relative_change", "absolute_change"],
    )
    # Plot relative change values
    ax = df_reshape["relative_change"].plot.bar(
        figsize=(20, 10),
        fontsize=16,
    )
    ax.set_xlabel("Time [y]", fontsize=20)
    ax.set_ylabel("[%]", fontsize=20)
    if len(df["year_range_lower"].unique()) > 1:
        add_title = " (wrt average of 5 previous years)"
    else:
        add_title = f" (wrt average of {df['year_range_lower'].unique()[0]}-{df['year_range_upper'].unique()[0]})"
    ax.set_title("Relative change" + add_title, fontsize=20)
    # Plot absolute change values
    ax = df_reshape["absolute_change"].plot.line(
        figsize=(20, 10),
        fontsize=16,
    )
    ax.set_xlabel("Time [y]", fontsize=20)
    if "unit" in group_data_pivot_table:
        ax.set_ylabel(
            f"[{'/'.join(df['unit'].unique().tolist())}]", fontsize=20
        )
    ax.set_title("Absolute change" + add_title, fontsize=20)
    # Show plot
    plt.show()


def plot_segmented_regression(df):
    """
    Function to plot segments defined by the linear regression interpolant (see segmented_regression function for the calculations)
    for each group of data
    :param (DataFrame) df, the output of segmented_regression function

    For example, select production quantity of palm oil from 1986 with Malaysia as reporter, compute the segmented regression analysis and plot results

        >>> from biotrade.faostat import faostat
        >>> from biotrade.common.time_series import segmented_regression, plot_segmented_regression
        >>> palm_oil_data = faostat.db.select(
        >>>     table="crop_production", reporter="Malaysia", product_code=257
        >>> )
        >>> palm_oil_prod = palm_oil_data[
        >>>     (palm_oil_data["element"] == "production") & (palm_oil_data["year"] > 1986)
        >>> ]
        >>> df_segment_reg = segmented_regression(palm_oil_prod, last_value=False, function="R2")
        >>> plot_segmented_regression(df_segment_reg)

    """
    # Define value column name
    if "value" in df.columns:
        # Value column of faostat db
        value_column = "value"
    elif "net_weight" in df.columns:
        # Value column is called "net_weight" in comtrade db
        value_column = "net_weight"
    # Define group columns
    groupby_column_list = ["reporter", "product_code", "unit"]
    # Depending on the db tables, different columns can be used to group the dataframe
    add_columns = ["partner", "element", "flow", "unit_code"]
    # Check if columns are present in the original dataframe
    for column in add_columns:
        if column in df.columns:
            # For comtrade db, do not consider unit and unit_code columns which are always nan
            if column == "unit_code":
                groupby_column_list.remove("unit")
                continue
            groupby_column_list.append(column)
    # Define groups with respect to calculate the segmented regressions
    df_groups = df.groupby(groupby_column_list)
    # Extract the information of significance level imposed and the type of obj function
    alpha = df["significance_level"].unique()[0]
    function = df["obj_function"].unique()[0]
    # Loop on each group to plot
    for key in df_groups.groups.keys():
        # Select data related to the specific group
        df_key = df_groups.get_group(key).sort_values(by="year")
        # Create arrays of independent (column "year") and dependent (colum "values") variables
        x = np.array(df_key["year"].values.tolist())
        y = np.array(df_key[value_column].values.tolist())
        # Plot the x-y scattered values
        plt.figure(figsize=(20, 10))
        plt.rc("font", size=16)
        plt.scatter(x, y, c="blue", s=50, label="Time series")
        plt.xlabel("Time [y]", fontsize=20)
        if "partner" in df_key:
            title = f"Reporter {df_key['reporter'].unique()[0]} - Partner {df_key['partner'].unique()[0]} ({function})"
        # No partner, only production
        else:
            title = f"Reporter {df_key['reporter'].unique()[0]} ({function})"
        plt.title(
            title,
            fontsize=20,
        )
        # Faostat tables
        if "element" in df_key:
            ylabel = f"Product {df_key['product'].unique()[0]} ({df_key['element'].unique()[0]}) [{df_key['unit'].unique()[0]}]"
        # Comtrade tables
        else:
            ylabel = f"Product {df_key['product'].unique()[0]} ({df_key['flow'].unique()[0]})"
        plt.ylabel(
            ylabel,
            fontsize=20,
        )
        # Plot segments
        for year_range in (
            df_key[["year_range_lower", "year_range_upper"]]
            .drop_duplicates()
            .values
        ):
            # Extract segment dataframe
            df_key_segment = df_key[
                (df_key["year"] >= year_range[0])
                & (df_key["year"] <= year_range[1])
            ]
            # Extract information to plot
            pvalue_segment = df_key_segment["pvalue"].unique()[0]
            lower_year_segment = df_key_segment["year_range_lower"].unique()[0]
            upper_year_segment = df_key_segment["year_range_upper"].unique()[0]
            slope_segment = df_key_segment["slope"].unique()[0]
            intercept_segment = df_key_segment["intercept"].unique()[0]
            # Plot only statically significant results below the significance level for the segmented regression
            if pvalue_segment < alpha:
                x_interval = np.array([lower_year_segment, upper_year_segment])
                y_interval = slope_segment * x_interval + intercept_segment
                plt.plot(
                    x_interval, y_interval, "ro-", label="Segmented regression"
                )
            # For the last segment the Mann Kendall test has been performed
            if upper_year_segment == x.max():
                mk_ha_segment = df_key_segment["mk_ha_test"].unique()[0]
                # If the test succeeded, plot the associated results
                if mk_ha_segment:
                    mk_slope_segment = df_key_segment["mk_slope"].unique()[0]
                    mk_intercept_segment = df_key_segment[
                        "mk_intercept"
                    ].unique()[0]
                    x_mk_interval = np.array(
                        [lower_year_segment, upper_year_segment]
                    )
                    y_mk_interval = (
                        mk_slope_segment * x_mk_interval + mk_intercept_segment
                    )
                    plt.plot(
                        x_mk_interval,
                        y_mk_interval,
                        "o-",
                        color="orange",
                        label="Theil–Sen estimator",
                    )
            # Do not plot vertical line if it is the last value of the time series
            else:
                plt.axvline(
                    x=upper_year_segment + 0.5,
                    linestyle="--",
                    color="k",
                    label="Breakpoint",
                )
        # Create legend without repetitions
        handles, labels = plt.gca().get_legend_handles_labels()
        by_label = OrderedDict(zip(labels, handles))
        plt.legend(by_label.values(), by_label.keys())

    # Show plot
    plt.show()


def obj_function(
    breakpoints, x, y, num_breakpoints, function, min_data_points, fcache
):
    """
    Function which computes the obj of the segmented regressions
    based on the mean of coefficient of determination (R2) with respect to the number of break points or based on the Residual Sum of Squares (RSS),
    depending on user choice.
    :param (np.array) breakpoints, the location of breakpoints
    :param (float) x, values of independent variable
    :param (float) y, values of dependent variable
    :param (string) function, the objective to calculate could be "R2" or "RSS"
    :param (integer) num_breakpoints, number of breaks, in other terms nr. segments - 1
    :param (string) function, the objective to calculate could be "R2" or "RSS", i.e. coefficient of determination based or
        residual sum of squares based (default = "RSS")
    :param (int) min_data_points, minimum nr of data for each segment of the linear regression (default = 7)
    :param (dictionary) fcache, values of the objective function for each evaluated
        location

    :return (dictionary) fcache updated
    """
    # Reorder breakpoint locations into a tuple
    breakpoints = tuple(map(int, sorted(breakpoints)))
    # Calculate obj function if not already done for the specific breakpoint location
    if breakpoints not in fcache:
        # Calculate the linear regression coefficients for the segments cut by breakpoint locations
        result = find_best_piecewise_polynomial(
            breakpoints, x, y, min_data_points
        )
        # Segments with more than 6 point
        if result:
            # Initialization
            obj = 0
            # Loop on each segment to obtain the total obj value
            for f, xi, yi in result:
                # Calculate the obj function as Residual Sum of Squares (RSS)
                if function == "RSS":
                    obj += np.sum((yi - (f.slope * xi + f.intercept)) ** 2)
                # Calculate the obj function based on coefficient of determination (R2) averaged over the number of breakpoints
                # Take the opposite value because obj needs to be minimize
                elif function == "R2":
                    if np.sum((yi - yi.mean()) ** 2) > 0:
                        obj += -(
                            1
                            - (np.sum((yi - (f.slope * xi + f.intercept)) ** 2))
                            / (np.sum((yi - yi.mean()) ** 2))
                        ) / (num_breakpoints + 1)
                    # If mean of values is 0, do not divide by it
                    else:
                        obj += -(
                            1
                            - (np.sum((yi - (f.slope * xi + f.intercept)) ** 2))
                        ) / (num_breakpoints + 1)
        # Penalize segments with less than min_data_points
        else:
            obj = np.inf
        fcache[breakpoints] = obj
    return fcache[breakpoints]


def find_best_piecewise_polynomial(
    breakpoints,
    x,
    y,
    min_data_points,
):
    """
    Function with computes the linear regression statistics for the breakpoint locations
    :param (np.array) breakpoints, the location of breakpoints
    :param (float) x, values of independent variable
    :param (float) y, values of dependent variable
    :param (int) min_data_points, minimum nr of data for each segment of the linear regression (default = 7)

    :return (list) result, containing statistics of the linear regression and values of the independent and
    dependent piece variables
    """
    # If the breakpoint is 1 transform it into a tuple
    if isinstance(breakpoints, np.floating):
        breakpoints = (int(breakpoints),)
    # Reorder breakpoint locations into a tuple
    else:
        breakpoints = tuple(map(int, sorted(breakpoints)))
    # Split x and y vectors into num_breakpoints +1 segments
    xs = np.split(x, breakpoints)
    ys = np.split(y, breakpoints)
    # Initialization
    result = []
    # Calculate the linear regression coefficients and store them into the result list
    for xi, yi in zip(xs, ys):
        # Do not consider empty arrays
        if not xi.size:
            continue
        # Do not consider linear regressions with less than minimum number of points argument
        elif 0 < len(xi) < min_data_points:
            # Break the loop and penalize the result putting an empty list
            result = []
            break
        # Compute the regression and store the results for each segment
        # https://docs.scipy.org/doc/scipy/reference/generated/scipy.stats.linregress.html
        f = stats.linregress(xi, yi)
        result.append([f, xi, yi])
    return result


def optimize_breakpoints(
    df, value_column, min_data_points, alpha, last_value, function
):
    """
    It allows to find the best number of breakpoints for a given time series and performs statistical tests
    on the optimized process

    :param (DataFrame) df, dataframe containing 1 time series for the optimization process
    :param (string) value_column, name of the column containing the data
    :param (int) min_data_points, minimum nr of data for each segment of the linear regression
    :param (float) alpha, significance level of statistical tests
    :param (boolean) last_value, if True returns only values related to the most recent year
    :param (string) function, the objective to calculate could be "R2" or "RSS", i.e. coefficient of determination or
        residual sum of squares

    :return (DataFrame) df, dataframe which contains new columns related to
    linear regression calculations for the given time series
    """
    # Avoid unexpected behavior of pandas apply function on df
    df = df.copy()
    # Select data related to the specific group dropping nan values
    df = df.sort_values(by="year").dropna(subset=value_column)
    # Skip segmented analysis for groups with nr of data less than minimum required
    if len(df) >= min_data_points:
        # Create arrays of independent (column "year") and dependent (colum "values") variables
        x = np.array(df["year"].values.tolist(), dtype=float)
        y = np.array(df[value_column].values.tolist())
        # Pre allocation
        best_breakpoints = ()
        # Loop over the number of breakpoints (max number of breakpoints is floor(len(x)/min_data_points) -1)
        # to check the best objective function result
        for num_breakpoints in range(0, math.floor(len(x) / min_data_points)):
            # If no breakpoints, no need to optimize the breakpoint location
            if num_breakpoints == 0:
                # Fake position
                breakpoints_pos = (0,)
                # Calculate the objective function for the entire array
                breakpoints_result = obj_function(
                    (0,), x, y, num_breakpoints, function, min_data_points, {}
                )
                # Store results
                breakpoints = (breakpoints_pos, breakpoints_result)
            # Call the optimization function optimize.brute which look at
            # a global minimum, moving 1 position at a time
            else:
                # Define the grid
                grid = [slice(1, len(x), 1)] * num_breakpoints
                # Return the optimize breakpoint location function
                # https://docs.scipy.org/doc/scipy/reference/generated/scipy.optimize.brute.html
                breakpoints = optimize.brute(
                    obj_function,
                    grid,
                    args=(x, y, num_breakpoints, function, min_data_points, {}),
                    full_output=True,
                    finish=None,
                )
            # Choose the best breakpoint locations considering also different number of breakpoints
            if best_breakpoints == ():
                best_breakpoints = breakpoints
            elif breakpoints[1] < best_breakpoints[1]:
                best_breakpoints = breakpoints
        # Add new columns to df reporting the linear regression statistics
        for f, xi, yi in find_best_piecewise_polynomial(
            best_breakpoints[0], x, y, min_data_points
        ):
            df.loc[
                (df["year"] >= xi.min()) & (df["year"] <= xi.max()),
                [
                    "slope",
                    "intercept",
                    "rsquared",
                    "pvalue",
                    "stderr_slope",
                    "stderr_intercept",
                    "year_range_lower",
                    "year_range_upper",
                ],
            ] = [
                f.slope,
                f.intercept,
                f.rvalue**2,
                f.pvalue,
                f.stderr,
                f.intercept_stderr,
                xi.min(),
                xi.max(),
            ]
            if xi.max() == x.max():
                # Significance level alpha (0.05 is the default)
                mk_results = mk.original_test(yi, alpha)
                # The estimate of slope is divided by the mean of time lags
                mk_results = mk_results._replace(
                    slope=mk_results.slope / np.mean(np.diff(xi))
                )
                # The estimate of the intercept is calculated by use of the Conover (1980) equation
                mk_results = mk_results._replace(
                    intercept=np.median(yi) - np.median(xi) * mk_results.slope
                )
                # Add the Mann-Kendall test results to the dataframe
                df.loc[
                    (df["year"] >= xi.min()) & (df["year"] <= xi.max()),
                    [
                        "mk_trend",
                        "mk_ha_test",
                        "mk_pvalue",
                        "mk_slope",
                        "mk_intercept",
                    ],
                ] = [
                    mk_results.trend,
                    mk_results.h,
                    mk_results.p,
                    mk_results.slope,
                    mk_results.intercept,
                ]
        # Return values related to most recent year at disposal
        if last_value:
            df = df.sort_values(by="year", ascending=False)[0:1]
    else:
        df = pd.DataFrame(
            columns=[
                *df.columns,
                "slope",
                "intercept",
                "rsquared",
                "pvalue",
                "stderr_slope",
                "stderr_intercept",
                "year_range_lower",
                "year_range_upper",
                "mk_trend",
                "mk_ha_test",
                "mk_pvalue",
                "mk_slope",
                "mk_intercept",
            ]
        )
    return df


def segmented_regression(
    df,
    last_value=True,
    function="RSS",
    min_data_points=7,
    alpha=0.05,
    multi_process=False,
    groupby_column_list=[],
    value_column=None,
):
    """
    Calculate the linear regression statistics of the segmented regression
    based on the optimization of the objective function

    :param (DataFrame) df, dataframe containing the time series with respect to calculate the segmented regressions
    :param (boolean) last_value, if True returns only values related to the most recent year (default = True)
    :param (string) function, the objective to calculate could be "R2" or "RSS", i.e. coefficient of determination based or
        residual sum of squares based (default = "RSS")
    :param (int) min_data_points, minimum nr of data for each segment of the linear regression (default = 7)
    :param (float) alpha, significance level of statistical tests (default = 0.05)
    :param multi_process (Boolean), if True segmented regression is performed through multiple cores. Default is False
    :param groupby_column_list (List), columns to be grouped. Default is empty list
    :param value_column (string), column which refers to data. Default is None

    :return (DataFrame) df_segmented_regression, dataframe which contains new columns related to
    linear regression calculations

    As example, compute the segmented regression for soybean exports from Brazil as reporter to Europe and
    rest of the world as partners.
    Import dependencies

        >>> from biotrade.faostat import faostat
        >>> from biotrade.faostat.aggregate import agg_trade_eu_row
        >>> from biotrade.common.time_series import segmented_regression

    Select soybean trade products with Brazil as reporter

        >>> soybean_trade = faostat.db.select(table="crop_trade",
        >>>     reporter="Brazil", product_code=236)

    Aggregate partners as Europe (eu) and rest of the World (row)

        >>> soybean_trade_agg = agg_trade_eu_row(soybean_trade)

    Select export quantity

        >>> soybean_exp_agg = soybean_trade_agg[soybean_trade_agg["element"] == "export_quantity"]

    Calculate the segmented regression of values returning the most recent year, first using RSS and then R2 objective function, keeping the minimum number of points and the significance level as default

        >>> soybean_exp_rss = segmented_regression(soybean_exp_agg, last_value = True, function = "RSS")
        >>> soybean_exp_r2 = segmented_regression(soybean_exp_agg, last_value = True, function = "R2")

    """

    # Define value column name if not provided
    if value_column is None:
        if "value" in df.columns:
            # Value column of faostat db
            value_column = "value"
        elif "net_weight" in df.columns:
            # Value column is called "net_weight" in comtrade db
            value_column = "net_weight"
        else:
            raise ValueError("Provide column name containing values")
    # Define group columns if not provided
    if len(groupby_column_list) == 0:
        groupby_column_list = ["reporter", "product_code", "unit"]
        # Depending on the db tables, different columns can be used to group the dataframe
        add_columns = ["source", "partner", "element", "flow", "unit_code"]
        # Check if columns are present in the original dataframe
        for column in add_columns:
            if column in df.columns:
                # For comtrade db, do not consider unit and unit_code columns which are always nan
                if column == "unit_code":
                    groupby_column_list.remove("unit")
                    continue
                groupby_column_list.append(column)
    # Group by df according to columns previously defined
    groupby = df.groupby(
        groupby_column_list,
        as_index=False,
    )
    # Define the function and arguments to be applied
    func = partial(
        optimize_breakpoints,
        value_column=value_column,
        min_data_points=min_data_points,
        alpha=alpha,
        last_value=last_value,
        function=function,
    )
    # Perform multiprocessing
    if multi_process:
        # Define data groups
        groups = [groupby.get_group(group) for group in groupby.groups]
        with Pool() as pool:
            # Launch parallel jobs using several cores depending on the machine
            segmented_regression = pool.map(
                func,
                groups,
            )
            # Concat results of the multi processing
            df_segmented_regression = pd.concat(
                segmented_regression, ignore_index=True
            )
    else:
        # Use the pd.apply function
        df_segmented_regression = groupby.apply(func).reset_index(drop=True)
    # Transform year lower and upper columns into int type
    df_segmented_regression[
        ["year_range_lower", "year_range_upper"]
    ] = df_segmented_regression[
        ["year_range_lower", "year_range_upper"]
    ].astype(
        "int"
    )
    # Add information regarding the obj function used
    df_segmented_regression["obj_function"] = function
    # Add information regarding the significance level
    df_segmented_regression["significance_level"] = alpha
    # Final dataframe with the new column statistic columns
    return df_segmented_regression


def calculate_changes(
    df, column_list_calc, year_range, value_column, years, last_value
):
    """
    Function to calculate absolute and relative changes

    :param (DataFrame) df, dataframe containing 1 time series for the change calculations
    :param (list) column_list_calc, column name list on which performing calculations
    ::param (list) year_range -> [lower year (int), upper year (int)], if values are given the function computes the average value with respect to this range and does not consider the years argument
    :param (string) value_column, name of the column containing the data
    :param (int) years, temporal window over calculating the relative and absolute change
    :param (boolean) last_value, if True returns only values related to the most recent year

    :return (DataFrame) df, dataframe which contains new columns related to
    change calculations for the given time series
    """
    # Avoid unexpected behavior of pandas apply function on df
    df = df.copy()
    # Fill nan year values
    dict_values = {}
    merge_col_list = []
    for col in df.columns:
        if col in ["year", "period"]:
            dict_values.update(
                {
                    col: range(
                        sorted(df.year.unique())[0],
                        sorted(df.year.unique())[-1] + 1,
                    )
                }
            )
            merge_col_list.append(col)
        elif col in [
            "source",
            "reporter_code",
            "reporter",
            "partner_code",
            "partner",
            "product_code",
            "product",
            "element_code",
            "element",
            "flow",
            "unit",
            "unit_code",
        ]:
            dict_values.update({col: df[col].unique()[0]})
            merge_col_list.append(col)
    df_fill_nan = pd.DataFrame(dict_values)
    df = df.merge(df_fill_nan, on=merge_col_list, how="outer")
    # Sort data from the most recent year
    df = df.sort_values(by="year", ascending=False)
    # Add columns to dataframe
    for idx, column in enumerate(column_list_calc):
        # The last column of the list corresponds to the average value
        if column == "average_value":
            # The last column is the mean of the corresponding values of the year range
            if year_range:
                df[column] = df.loc[
                    (df["year"] >= year_range[0])
                    & (df["year"] <= year_range[1]),
                    value_column,
                ].mean()
            # The last column is the mean of the previous year values
            elif years > 0:
                df[column] = df[column_list_calc[:-1]].mean(
                    axis=1,
                )
            break
        # Shift value column up with respect to the number of years considered
        df[column] = df[value_column].shift(-(idx) - 1)
    # Finally calculate the relative and absolute change columns
    df["relative_change"] = (
        (df[value_column] - df["average_value"]) / df["average_value"] * 100
    )
    df["absolute_change"] = df[value_column] - df["average_value"]
    # 0/0 and float/0 treated as NaN
    df["relative_change"].replace([np.inf, -np.inf], np.nan, inplace=True)
    # Define 2 new columns with the lower and upper years of the average value
    if year_range:
        df["year_range_lower"] = year_range[0]
        df["year_range_upper"] = year_range[1]
    elif years > 0:
        df["year_range_lower"] = df["year"] - years
        df["year_range_upper"] = df["year"] - 1
    # Drop columns added for the calculation
    df = df.drop(column_list_calc[:-1], axis=1)
    # Return values related to most recent year at disposal
    if last_value:
        df = df[0:1]

    return df


def relative_absolute_change(
    df,
    years=5,
    last_value=True,
    multi_process=False,
    year_range=[],
    groupby_column_list=[],
    value_column=None,
):
    """
    Calculate the relative and absolute change of a quantity with respect to
    values assumed by that quantity in the n previous years or with respect to a year range given, for
    production and trade data.
    Relative change of a value is calculated as:
        (value of the reference year - mean of values of previous n years (or year range given)) / mean of values of previous n years (or year range given) * 100
    Absolute change of a value is calculated as:
        value of the reference year - mean of values of previous n years (or year range given)

    :param (DataFrame) df, dataframe containing the quantity with respect to calculate the relative and absolute change in time
    :param (int) years, temporal window over calculating the relative and absolute change (default = 5)
    :param (boolean) last_value, if True returns only values related to the most recent year (default = True)
    :param (list) year_range -> [lower year (int), upper year (int)], if given the function computes the average value with respect to this range and does not consider the years argument (default = empty)
    :param multi_process (Boolean), if True changes are performed through multiple cores. Default is False
    :param groupby_column_list (List), columns to be grouped. Default is empty list
    :param value_column (string), column which refers to data. Default is None

    :return (DataFrame) df_change, dataframe which contains new columns
        "average_value" mean of values of the year range
        "relative_change" value calculated
        "absolute_change" value calculated
        "year_range_lower" lower year with respect to compute the average value
        "year_range_upper" upper year with respect to compute the average value

    As example, compute the relative and absolute change for soy trade products from Brazil as reporter to Europe and
    rest of the world as partners.
    Import dependencies

        >>> from biotrade.faostat import faostat
        >>> from biotrade.faostat.aggregate import agg_trade_eu_row
        >>> from biotrade.common.time_series import relative_absolute_change

    Select soy trade products with Brazil as reporter

        >>> soy_trade = faostat.db.select(table="crop_trade",
        >>>     reporter="Brazil", product="soy")

    Aggregate partners as Europe (eu) and rest of the World (row)

        >>> soy_trade_agg = agg_trade_eu_row(soy_trade)

    Calculate the relative and absolute change of values in time for the most recent year, considering the last 5 years

        >>> soy_trade_change1 = relative_absolute_change(soy_trade_agg, years= 5, last_value = True)

    Calculate the relative and absolute change of values in time with respect to the average value between 2000 and 2010

        >>> soy_trade_change2 = relative_absolute_change(soy_trade_agg, year_range = [2000, 2010], last_value = False)

    """
    # Define value column name if not provided
    if value_column is None:
        if "value" in df.columns:
            # Value column of faostat db
            value_column = "value"
        elif "net_weight" in df.columns:
            # Value column is called "net_weight" in comtrade db
            value_column = "net_weight"
        else:
            raise ValueError("Provide column name containing values")
    # Define group columns if not provided
    if len(groupby_column_list) == 0:
        groupby_column_list = ["reporter", "product_code", "unit"]
        # Depending on the db tables, different columns can be used to group the dataframe
        add_columns = [
            "source",
            "partner",
            "element",
            "flow",
            "unit_code",
        ]
        # Check if columns are present in the original dataframe
        for column in add_columns:
            if column in df.columns:
                # For comtrade db, do not consider unit and unit_code columns which are always nan
                if column == "unit_code":
                    groupby_column_list.remove("unit")
                    continue
                groupby_column_list.append(column)
    # Define groups with respect to calculate the relative and absolute change of values in time
    groupby = df.groupby(groupby_column_list, as_index=False)
    # Allocate df reshaped to return
    df_change = pd.DataFrame()
    # Column list to perform calculations
    column_list_calc = []
    # If a year range is given, consider it to calculate the average value
    if year_range:
        years = 0
    # If a year range is not given, consider the n years before
    elif years > 0:
        # Add columns with respect to the number of years to calculate the relative and absolute change
        for year in range(1, years + 1):
            column_list_calc.append(f"value_{year}_year_before")
    # Colum to perform the average value
    column_list_calc.append("average_value")
    # Define the function and arguments to be applied
    func = partial(
        calculate_changes,
        column_list_calc=column_list_calc,
        year_range=year_range,
        value_column=value_column,
        years=years,
        last_value=last_value,
    )
    # Perform multiprocessing
    if multi_process:
        # Define data groups
        groups = [groupby.get_group(group) for group in groupby.groups]
        with Pool() as pool:
            # Launch parallel jobs using several cores depending on the machine
            relative_absolute = pool.map(
                func,
                groups,
            )
            # Concat results of the multi processing
            df_change = pd.concat(relative_absolute, ignore_index=True)
    else:
        # Use the pd.apply function
        df_change = groupby.apply(func).reset_index(drop=True)
    # Final dataframe with the new columns
    return df_change


def merge_analysis(df_change, df_segmented_regression):
    """
    Function to merge dataframes related to change and segmented regression analysis, in order to compare results.

    :param (DataFrame) df_change, df obtained from relative_absolute_change function
    :param (DataFrame) df_change, df obtained from segmented_regression function
    :return (DataFrame) df, which is a merge of the input dataframes

    For example select soybean trade product with Brazil as reporter, aggregate partners as Europe (eu) and rest of the World (row)
    and compute the change and segmented analysis. Finally compare results with the merge function

    >>> from biotrade.faostat import faostat
    >>> from biotrade.faostat.aggregate import agg_trade_eu_row
    >>> from biotrade.common.time_series import (relative_absolute_change, segmented_regression, merge_analysis)
    >>> soybean_trade = faostat.db.select( table="crop_trade", reporter="Brazil", product_code=236)
    >>> soybean_trade_agg = agg_trade_eu_row(soybean_trade)
    >>> soybeans_exp_change = relative_absolute_change(soybean_trade_agg, last_value=True)
    >>> soybean_exp_regression = segmented_regression(soybean_trade_agg, last_value=True, function="R2", alpha=0.05, min_data_points=7)
    >>> df_merge = merge_analysis(df_change=soybeans_exp_change, df_segmented_regression=soybean_exp_regression)

    """
    # Columns in common for the two dataframes
    join_columns = list(
        set(df_change.columns) & set(df_segmented_regression.columns)
    )
    # Columns not to be considered for the join
    remove_columns = ["year_range_lower", "year_range_upper"]
    # Final join columns
    join_columns = list(set(join_columns) - set(remove_columns))
    # Merge the two dataframes
    df = df_change.merge(
        df_segmented_regression,
        on=join_columns,
        how="outer",
        suffixes=("_change", "_regression"),
    )
    # Columns to be retained
    df = df[
        [
            *join_columns,
            "relative_change",
            "absolute_change",
            "year_range_lower_change",
            "year_range_upper_change",
            "mk_slope",
            "mk_ha_test",
            "year_range_lower_regression",
            "year_range_upper_regression",
        ]
    ]

    return df



from biotrade.comtrade import comtrade
from biotrade.faostat import faostat
from biotrade.world_bank import world_bank


def cron_quarterly():
    """Update of yearly datasets every three months"""
    comtrade.pump.update_db(table_name="yearly", frequency="A", start_year=2000)
    comtrade.pump.update_db_parameter()
    faostat_tables = [
        "country",
        "crop_production",
        "crop_trade",
        "food_balance",
        "forestry_production",
        "forestry_trade",
        "forest_land",
        "land_use",
        "land_cover",
    ]
    faostat.pump.update(faostat_tables, skip_confirmation=True)
    world_bank.pump.update(["indicator", "indicator_name"], skip_confirmation=True)


def cron_bi_weekly():
    """Update of monthly datasets every two weeks"""
    comtrade.pump.update_db(table_name="monthly", frequency="M")


def cron_front_end():
    """Update of web platform datasets every six months"""
    from scripts.front_end import product_list
    from scripts.front_end import annual_variation_production
    from scripts.front_end import average_production
    from scripts.front_end import trends_production
    from scripts.front_end import annual_variation_trade
    from scripts.front_end import average_trade
    from scripts.front_end import statement

    product_list.main()
    annual_variation_production.main()
    average_production.main()
    trends_production.main()
    annual_variation_trade.main()
    average_trade.main()
    statement.main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Written by Paul Rougieux and Selene Patani.

Copyright (c) 2023 European Union
Licenced under the MIT licence

JRC biomass Project.
Unit D1 Bioeconomy.


The database can be updated every 2 weeks with a cron job such as

To run comtrade.pump.update_db periodically as a cron job, edit crontab:

    sudo vim /etc/crontab

And enter:

    0 0 0 */2 * cronjob_update_db.py

Where the cronjob_update_db.py file contains

    from biotrade.comtrade import comtrade
    comtrade.pump.update_db(table_name = "monthly", frequency = "M")

"""
# Built-in modules
from pathlib import Path
import warnings
import tempfile
import shutil
from zipfile import ZipFile
import datetime
import pytz
import os
import re
import time
from collections import OrderedDict

try:
    import requests
except Exception as e:
    msg = "Failed to import requests, you will not be able to load data from Comtrade,"
    msg += "but you can still use other methods.\n"
    print(msg, str(e))

# Third party modules
import json
import logging
import pandas
import numpy as np

try:
    import comtradeapicall
except Exception as e:
    msg = "Failed to import comtradeapicall package, you will not be able to load data from there,"
    msg += "but you can still use other methods.\n"
    print(msg, str(e))

# Internal modules
from biotrade.common.url_request_header import HEADER


class Pump:
    """
    Download trade data from the Comtrade API and store it inside a database.

    There are two different ways to download data from the Comtrade API:
        - the public API
        - a restricted bulk API accessible with a token

    The public API is accessible to every one and is suitable to download small
    amounts of data for a few products in a few countries. For example, the
    following calls download trade data for the selected product code `cc`,
    reporter `r` and time period `ps` from the Comtrade API and returns data
    frames that contain sawnwood other (HS code 440799) and soya beans (HS
    code 120190) trade reported by Italy in 2020:

        >>> from biotrade.comtrade import comtrade
        >>> # Other sawnwood
        >>> sawnwood99 = comtrade.pump.download_df(cc = "440799", r="381", ps="2020")
        >>> # Soy
        >>> soya_beans = comtrade.pump.download_df(cc = "120190", r="381", ps="2020")

    The public API can also be used to get the list of products and product
    codes:

        >>> hs = comtrade.pump.get_parameter_list("classificationHS.json")

    Lists of countries and country codes are also available:

        >>> comtrade.pump.get_parameter_list("reporterAreas.json")
        >>> comtrade.pump.get_parameter_list("partnerAreas.json")

    In case you have access to the API token. It's also possible to download
    data from the Comtrade bulk API and store it in the database with:

        >>> from biotrade.comtrade import comtrade
        >>> comtrade.pump.update_db(table_name = "monthly", frequency = "M")

    Potential Error messages returned by the Comtrade API:

    - HTTP Error 401 "Unauthorized" means that the token authentication is
      required and missing.
    - HTTP Error 409 "Conflict" seems to be the error when the download limit has been reached
      and we don't use the authentication token.
    - HTTP Error 500 "Internal Server Error" appears when the max argument is
      greater than 9999.

    """

    # Log debug and error messages
    logger = logging.getLogger("biotrade.comtrade")
    # Define URL request headers
    header = HEADER
    # Base URL to load trade trade data from the API
    url_api_base = "https://comtrade.un.org/api/get"
    # Base URL to load metadata
    url_metadata_base = "https://comtrade.un.org/Data/cache/"

    def __init__(self, parent):
        # Default attributes #
        self.parent = parent
        self.db = self.parent.db
        # Mapping table used to rename columns
        self.column_names = self.parent.column_names
        # Get API authentication token from an environmental variable
        self.token = None
        if os.environ.get("COMTRADE_TOKEN"):
            self.token = os.environ["COMTRADE_TOKEN"]
        # Path of CSV log file storing API parameters and download status
        self.csv_log_path = self.parent.data_dir / "pump_comtrade_api_args.csv"

    def sanitize_variable_names(self, df, renaming_from, renaming_to):
        """
        Sanitize column names using the mapping table column_names.csv
        Use snake case in column names and replace some symbols
        """
        # Rename columns to snake case
        df.rename(columns=lambda x: re.sub(r" ", "_", str(x)).lower(), inplace=True)
        # Remove parenthesis and dots, used only for human readable dataset
        df.rename(columns=lambda x: re.sub(r"[()\.]", "", x), inplace=True)
        # Replace $ sign by d, used only for human readable dataset
        df.rename(columns=lambda x: re.sub(r"\$", "d", x), inplace=True)
        # Rename columns using the naming convention defined in self.column_names
        mapping = self.column_names.set_index(renaming_from).to_dict()[renaming_to]
        # Discard nan keys of mapping dictionary
        mapping.pop(np.nan, None)
        df.rename(columns=mapping, inplace=True)
        # Rename column content to snake case using a compiled regex
        regex_pat = re.compile(r"\W+")
        if "flow" in df.columns:
            df["flow"] = df["flow"].str.replace(regex_pat, "_", regex=True).str.lower()
            # Remove the plural "s"
            df["flow"] = df["flow"].str.replace("s", "", regex=True)
        return df

    def download_bulk_gz(self, period, frequency):
        """
        Method of Pump class to download bulk data from comtradeapicall package, which returns country gz files
        stored into a temporary directory

        :param (int) period, as year or year+month
        :param (str) frequency, as "A" yearly or "M" monthly
        :output (path) temp_dir, directory path which contains the gz files
        :output (int) response_code, response status of the comtradeapicall

        For example monthly bilateral trade gz files for all products in November
        2021 can be downloaded into a temporary directory as:

            >>> from biotrade.comtrade import comtrade
            >>> temp_dir, resp_code = comtrade.pump.download_bulk_gz(202111, "M")
        """

        # Temporary folder creation
        temp_dir = Path(tempfile.mkdtemp())
        self.logger.info(
            "Downloading %s data for period %s from:\n %s.%s",
            "yearly" if frequency == "A" else "monthly",
            period,
            comtradeapicall.__name__,
            comtradeapicall.bulkDownloadFinalFile.__name__,
        )
        # Variable for checking successful gz download
        download_successful = False
        # Default time to download gz files
        time_start = time.time()
        sleep_time = 0
        # Send the request: if successful put the gz files into the temporary
        # folder else retry.
        # If the wait raises to more than 1 hour the loop stops.
        while not download_successful and sleep_time < 3600:
            try:
                # Create a request
                comtradeapicall.bulkDownloadFinalFile(
                    self.token,
                    temp_dir,
                    typeCode="C",
                    freqCode=frequency,
                    clCode="HS",
                    period=period,
                    decompress=False,
                )
                response_code = 200
                download_successful = True
                # No data available
                if len(os.listdir(temp_dir)) == 0:
                    response_code = 404
            except (Exception, BaseException) as error:
                response_code = error
            sleep_time = time.time() - time_start
            self.logger.info(f"HTTP response code: {response_code}")
        return temp_dir, response_code

    def download_bulk_csv(self, period, frequency):
        """
        Method of Pump class to download bulk data from API
        https://comtrade.un.org/data/doc/api/bulk/ which returns a zip file
        stored into a temporary directory

        :param (int) period, as year or year+month
        :param (str) frequency, as "A" yearly or "M" monthly
        :output (path) temp_dir, directory path which contains the zip file
        :output (int) response_code, response status of the API request

        For example monthly bilateral trade data for all products in November
        2021 can be downloaded into a temporary directory as:

            >>> from biotrade.comtrade import comtrade
            >>> temp_dir, resp_code = comtrade.pump.download_bulk_csv(202111, "M")
        """

        # Construnction of API bulk url
        url_api_call = (
            f"{self.url_api_base}/bulk/C/{frequency}/{period}/ALL/HS?"
            + f"token={self.token}"
        )
        # Temporary folder creation
        temp_dir = Path(tempfile.mkdtemp())
        # Temporary zip file path
        temp_file = temp_dir / "temp.zip"
        self.logger.info("Downloading data from:\n %s", url_api_call)
        # Variable for checking successfull zip dowload
        download_successful = False
        # Default sleep time before downloading zip file
        sleep_time = 5
        # Send the request: if successful put the zip file into the temporary
        # folder else retry and double the wait time at each try
        # If the wait raises to more than 1 hour the loop stops.
        while not download_successful and sleep_time < 3600:
            time.sleep(sleep_time)
            try:
                # Create a request
                response = requests.get(
                    url=url_api_call, headers=self.header, stream=True
                )
                # Raise in case of HTTP error
                response.raise_for_status()
                with open(temp_file, "wb") as out_file:
                    # Copy the zip file into the output_file
                    shutil.copyfileobj(response.raw, out_file)
                    download_successful = True
                    response_code = response.status_code
            except requests.exceptions.HTTPError as error:
                response_code = error.response.status_code
                # Error 404 means that no data are present, stop the while loop
                if response_code == 404:
                    download_successful = True
            except requests.exceptions.RequestException as error:
                response_code = error
            sleep_time *= 2
            self.logger.info(f"HTTP response code: {response_code}")
        return temp_dir, response_code

    def transfer_gz_files(
        self,
        temp_dir,
        table_name,
        bioeconomy_tuple,
        check_data_presence,
        api_period,
    ):
        """
        Pump method to transfer gz files to db using dataframe.
        500k data frame rows ~ 350 MB of memory usage
        4 millions data frame rows ~ 2.8 GB of memory usage

        :param (path) temp_dir, path of the folder containing the gz files
            to copy into db
        :param (str) table_name, table name of the db
        :param (tuple) bioeconomy_tuple, commodity codes to store data
        :param (bool) check_data_presence, to delete data in case of existing
            rows into db
        :param (int) api_period, period to delete existing data
        :return (int) records_transferred, rows uploaded to db
        """
        # List of chunk rows
        chunk_list = []
        # Read the gz files
        for temp_file in os.listdir(temp_dir):
            # Preallocate a dataframe
            df = pandas.DataFrame()
            nr_attempts = 0
            df_from_gz = False
            while not df_from_gz and nr_attempts < 10:
                # Loop on gz files and upload them
                try:
                    df = pandas.read_csv(
                        temp_dir / temp_file,
                        sep="\t",
                        dtype={
                            "datasetCode": str,
                            "cmdCode": str,
                            "mosCode": str,
                        },
                    )
                    df_from_gz = True
                # Continue until max number of attempts is reached
                except Exception as error_gz:
                    self.logger.info(
                        f"Unable to load data from {temp_file} for period {api_period} due to\n {error_gz}.\n Retrying to download it.."
                    )
                    try:
                        file_str = re.split(r"[-\[\]]", temp_file)[2]
                        frequency = file_str[1]
                        reporter_code = int(file_str[2:5])
                        period = int(file_str[5:-2])
                        comtradeapicall.bulkDownloadFinalFile(
                            self.token,
                            temp_dir,
                            typeCode="C",
                            freqCode=frequency,
                            clCode="HS",
                            period=period,
                            reporterCode=reporter_code,
                            decompress=False,
                        )
                        response_code = 200
                    except (Exception, BaseException) as error:
                        response_code = error
                    self.logger.info(f"Response: {response_code}")
                    nr_attempts += 1
            if nr_attempts == 10:
                self.logger.warning(
                    f"Reached the max number of attempts, {nr_attempts}. Unable to load data from {temp_file} for period {api_period}"
                )
            # If length is > 0 select rows
            if not df.empty:
                # Remove potential white spaces between tab delimiter in string columns
                strip_cols = df.select_dtypes(["object"]).columns
                df[strip_cols] = df[strip_cols].apply(lambda x: x.str.strip())
                nrow_before = len(df)
                memory_before = round(df.memory_usage(deep=True).sum() / (1024**2), 2)
                selector = (
                    (df["customsCode"] == "C00")
                    & (df["motCode"] == 0)
                    & (df["mosCode"] == "0")
                    & (df["partner2Code"] == 0)
                    & (df["flowCode"].isin(["M", "X", "RM", "RX"]))
                )
                if table_name in ("monthly", "yearly"):
                    # Store codes with bioeconomy label and 6 digits, not differentiating modality of transport (0), supply ("0"), 2nd partner (0)  and procedures ("C00")
                    # only (re)exported and (re)imported data
                    selector_6d = (
                        selector
                        & df["cmdCode"].str.startswith(bioeconomy_tuple)
                        & (df["cmdCode"].str.len() == 6)
                    )
                    df = df[selector_6d]
                elif table_name == "yearly_hs2":
                    # Store codes with bioeconomy label and 2 digits, not differentiating modality of transport (0), supply ("0"), 2nd partner (0)  and procedures ("C00")
                    # only (re)exported and (re)imported data
                    selector_2d = selector & df["cmdCode"].isin(bioeconomy_tuple)
                    df = df[selector_2d]
                msg = f"Before filtering {temp_file[:-3]} file, the dataframe occupies {memory_before} MB and number of rows are {nrow_before}, after {round(df.memory_usage(deep=True).sum() / (1024**2), 2)} MB with {len(df)} rows"
                print(msg)
                # Append rows
                chunk_list.append(df)
        # Construct the final df to upload to db
        if chunk_list:
            df = pandas.concat(chunk_list)
            self.logger.info(
                "Memory usage of the filtered dataframe corresponding to period %s:\n%s GB",
                api_period,
                round(df.memory_usage(deep=True).sum() / (1024**3), 2),
            )
        if not df.empty:
            # Call method to rename column names
            df = self.sanitize_variable_names(
                df, renaming_from="comtrade_apicall", renaming_to="biotrade"
            )
            # Delete already existing data
            if check_data_presence:
                self.db.delete_data(
                    table_name,
                    api_period,
                    api_period,
                )
            # Append data to db
            self.db.append(df, table_name)
        else:
            self.logger.warning(
                f"Dataframe for period {api_period} is empty. Previous data in the db are not deleted."
            )
        # Keep track of the length of the transferred data in the log file
        records_transferred = len(df)
        return records_transferred

    def transfer_csv_chunk(
        self,
        temp_file,
        table_name,
        bioeconomy_tuple,
        check_data_presence,
        api_period,
        chunk_size=10**6,
    ):
        """
        Pump method to transfer large csv file to db using dataframe.
        500k data frame rows ~ 500 MB of memory usage
        2 millions data frame rows ~ 2 GB of memory usage

        :param (path) temp_file, path of the zip file containing the csv file
            to copy into db
        :param (str) table_name, table name of the db
        :param (tuple) bioeconomy_tuple, commodity codes to store data
        :param (bool) check_data_presence, to delete data in case of existing
            rows into db
        :param (int) api_period, period to delete existing data
        :param (int) chunk_size, splitting csv to transfer to df default is 10**6
        """
        # Number of records transferred from csv file
        records_downloaded_csv = 0
        # Open the zip file
        with ZipFile(temp_file) as zipfile:
            with zipfile.open(zipfile.namelist()[0]) as csvfile:
                # List of chunk rows
                chunk_list = []
                # Loop on csv chunks and upload them
                for df_chunk in pandas.read_csv(
                    csvfile,
                    encoding="utf-8",
                    # Force the id column to remain a character column,
                    # otherwise str "01" becomes int 1.
                    dtype={"Commodity Code": str, "Reporter ISO": str},
                    chunksize=chunk_size,
                ):
                    # If length is > 0 select rows
                    if not df_chunk.empty:
                        if table_name == "monthly":
                            # Store codes with bioeconomy label and 6 digits
                            df_chunk = df_chunk[
                                (
                                    df_chunk["Commodity Code"].str.startswith(
                                        bioeconomy_tuple
                                    )
                                )
                                & (df_chunk["Commodity Code"].str.len() == 6)
                            ]
                        elif table_name == "yearly_hs2":
                            # Store codes with bioeconomy label and 2 digits
                            df_chunk = df_chunk[
                                df_chunk["Commodity Code"].isin(bioeconomy_tuple)
                            ]
                        elif table_name == "yearly":
                            # Store codes with bioeconomy label and 6 digits
                            df_chunk = df_chunk[
                                (
                                    df_chunk["Commodity Code"].str.startswith(
                                        bioeconomy_tuple
                                    )
                                )
                                & (df_chunk["Commodity Code"].str.len() == 6)
                            ]
                        # Append rows
                        chunk_list.append(df_chunk)
        # Construct the final df to upload to db
        df = pandas.concat(chunk_list)
        self.logger.info(
            "Memory usage:\n%s GB",
            round(df.memory_usage(deep=True).sum() / (1024**3), 2),
        )
        if not df.empty:
            # Call method to rename column names
            df = self.sanitize_variable_names(
                df, renaming_from="comtrade_human", renaming_to="biotrade"
            )
            # Delete already existing data
            if check_data_presence:
                self.db.delete_data(
                    table_name,
                    api_period,
                    api_period,
                )
            # Append data to db
            self.db.append(df, table_name)
        # Keep track of the the length of the data in the log file
        records_downloaded_csv += len(df)
        return records_downloaded_csv

    def transfer_bulk_csv(
        self,
        table_name,
        start_year,
        end_year,
        frequency,
        check_data_presence,
    ):
        """
        Pump method to transfer bulk files of Comtrade API requests/package to
        Comtrade db.
        Performance with Intel(R) Core(TM) i7-1065G7 CPU @ 1.30GHz:
        ~ 1.5 hours to upload db monthly data for 1 year
        ~ 9 hours to upload db monthly data for 6 years

        :param (str) table_name, name of the db table to store data
        :param (int) start_year, year from the download should start
        :param (int) start_year, year from the download should end
        :param (str) frequency, as "A" yearly or "M" monthly
        :param (bool) check_data_presence, if data already exists into db

        Example for uploading data from 2016 to 2017 with monthly data into
        "monthly" comtrade table:

        First check if data already exists into db

            >>> from biotrade.comtrade import comtrade
            >>> data_check = comtrade.db.check_data_presence(
            >>>     table = "monthly",
            >>>     start_year = 2016,
            >>>     end_year = 2017,
            >>>     frequency = "M",
            >>> )

        Upload data to db

            >>> comtrade.pump.transfer_bulk_csv(
            >>>     table_name = "monthly",
            >>>     start_year = 2016,
            >>>     end_year = 2017,
            >>>     frequency = "M",
            >>>     check_data_presence = data_check,
            >>> )

        """
        # Adjust frequency parameter in case it is wrong provided
        if table_name in ("yearly", "yearly_hs2"):
            frequency = "A"
        elif table_name == "monthly":
            frequency = "M"
        # Period list of downloads failed
        period_list_failed = []
        # Total records transferred from zip files of API requests
        total_records = 0
        # Download of bioeconomy codes from table "comtrade_hs_2d.csv"
        path = self.parent.config_data_dir / "comtrade_hs_2d.csv"
        # Store table codes into a data frame
        prod = pandas.read_csv(path)
        # Selection of codes which have bioeconomy column = 1
        bioeconomy_tuple = tuple(
            prod[prod.bioeconomy == 1]["product_description"].str[:2]
        )
        # Range of period to download data
        period_block = range(start_year, end_year + 1)
        # Add months for monthly frequency
        if frequency == "A":
            month_list = [""]
        elif frequency == "M":
            month_list = [
                "01",
                "02",
                "03",
                "04",
                "05",
                "06",
                "07",
                "08",
                "09",
                "10",
                "11",
                "12",
            ]
        # Date object of today
        current_date = datetime.datetime.now(pytz.timezone("Europe/Rome")).date()
        # Loop on year and eventually month, depending on the frequency
        # parameter
        for period in period_block:
            # Data not available in the future
            if period > current_date.year:
                break
            for month in month_list:
                if frequency == "M":
                    # Data not available in the future
                    if datetime.datetime(period, int(month), 1).date() > current_date:
                        break
                # Construct the period to pass to transfer_csv_chunk method
                api_period = int(str(period) + month)
                # Use the new Comtrade API package
                # Store gz data into the temporary directory
                temp_dir, response_code = self.download_bulk_gz(
                    api_period,
                    frequency,
                )
                # If data are downloaded (response = 200) copy data into a pandas data frame and upload it to the db
                if response_code == 200:
                    # Store into the database
                    try:
                        # Transfer gz files to db
                        records_downloaded = self.transfer_gz_files(
                            temp_dir,
                            table_name,
                            bioeconomy_tuple,
                            check_data_presence,
                            api_period,
                        )
                        # Total number of records
                        total_records += records_downloaded
                        self.logger.info(
                            f"Updated database table {table_name} for period"
                            + f" {api_period}\n"
                            + f"{total_records} records uploaded in total."
                        )
                    # Failed to store data into db
                    except Exception as error_db:
                        self.logger.warning(
                            "Failed to store data into the database for table"
                            + f" {table_name} and period {api_period}\n"
                            + f"{error_db}"
                        )
                        period_list_failed.append(api_period)
                # Data not downloaded from API requests/package
                else:
                    self.logger.warning(
                        "Failed to download from API request for"
                        + f" period {api_period} due to HTTP/URL error"
                    )
                    period_list_failed.append(api_period)
                # Remove temporary directory
                shutil.rmtree(temp_dir)
        # Log info if some periods failed to be uploaded into db
        if len(period_list_failed):
            self.logger.warning(
                "List of failed download periods for table"
                + f" {table_name}: {period_list_failed}"
            )

    def update_db(self, table_name, frequency, start_year=None):
        """
        Pump method to update db. If data from 2016 are already present,
        it updates data of the last and current year, otherwise it uploads
        data from 2016.

        :param (string) table_name, name of Comtrade db table
        :param (string) frequency, "M" for monthly data or "A" for annual
        :param (int) start_year, year to start the download from, defaults to
        last year if not specified

        Return an error if some periods are not uploaded to the database.

        Example of updating db table "monthly" with monthly frequency data:

            >>> from biotrade.comtrade import comtrade
            >>> comtrade.pump.update_db(table_name = "monthly", frequency = "M")

        """
        # Adjust frequency parameter in case it is wrongly provided
        if table_name in ("yearly", "yearly_hs2"):
            frequency = "A"
        elif table_name == "monthly":
            frequency = "M"
        current_year = datetime.datetime.now(pytz.timezone("Europe/Rome")).date().year
        # If start year not specified, update from the last year only
        if start_year is None:
            start_year = current_year - 1
        # Check if data from the start year are present into the database
        data_present = self.db.check_data_presence(
            table_name,
            start_year,
            current_year,
            frequency,
        )
        # Transfer from api bulk requests to db
        self.transfer_bulk_csv(
            table_name,
            start_year,
            current_year,
            frequency,
            data_present,
        )

    def download_df(
        self,
        max="500",
        type="C",
        freq="A",
        px="HS",
        ps="2020",
        r="381",
        p="all",
        rg="all",
        cc="440799",
        fmt="json",
        head="M",
    ):
        """Download a CSV file from the UN Comtrade data API and return a pandas data frame.

        The data API is documented at https://comtrade.un.org/data/doc/api/

        Argument names are the same as the Comtrade API argument names.

        :param str max : maximum number of rows returned by the API
         Note the server returns "HTTP Error 500: Internal Server Error" if
         max is greater than 9999.
        :param str type : type of trade (c = commodities, s = services)
        :param str freq : frequency A Annual, M Monthly
        :param str px : classification
        :param str ps : time period
        :param str r : reporter country
        :param str p : partner country
        :param str rg : trade flow
        :param str cc : product code
        :param str fmt : data format, defaults to "json",
                         usage of "csv" is discouraged as it can cause data type issues
        :param str head : human (H) or machine (M) readable column headers
        """
        if fmt not in ("csv", "json"):
            raise ValueError(f"The data format '{fmt}' is not supported")

        # Build the API call
        url_api_call = (
            f"{self.url_api_base}?max={max}&type={type}&freq={freq}&px={px}"
            + f"&ps={ps}&r={r}&p={p}&rg={rg}&cc={cc}&fmt={fmt}&head={head}"
        )
        if self.token is not None:
            url_api_call += f"&token={self.token}"
        self.logger.info("Downloading a data frame from:\n %s", url_api_call)

        # Load the data in csv format
        if fmt == "csv":
            df = pandas.read_csv(
                url_api_call,
                # Force the id column to remain a character column,
                # otherwise str "01" becomes int 1.
                dtype={"Commodity Code": str, "cmdcode": str},
            )

        # Load the data in json format
        if fmt == "json":
            response = requests.get(url=url_api_call, headers=self.header, stream=True)
            print(f"HTTP response code: {response.status_code}")
            json_content = json.load(response.raw)
            # If the data was downloaded incorrectly, raise an error with the validation status
            if json_content["validation"]["status"]["name"] != "Ok":
                raise requests.exceptions.RequestException(
                    f"API message\n{json_content['validation']}"
                )
            else:
                df = pandas.json_normalize(json_content["dataset"])

        # Raise an error if the data frame has the maximum number of rows returned by the query
        if len(df) >= int(max) - 1:
            raise ValueError(
                f"The number of rows in the returned data ({len(df)}) "
                + f"is equal to the maximum number of rows ({max}). "
                + f"Increase the max argument for product {cc} and reporter {r} "
                + "in order to get all requested data from Comtrade."
            )
        df = self.sanitize_variable_names(
            df, renaming_from="comtrade_machine", renaming_to="biotrade"
        )
        return df

    def get_parameter_list(self, file_name):
        """Get the list of valid parameters from the Comtrade API
        A list of parameter in json format are available and explained on
        https://comtrade.un.org/data/doc/api

        For example to get the list of reporters

        >>> from biotrade.comtrade import comtrade
        >>> comtrade.pump.get_parameter_list("reporterAreas.json")

        Get the list of partners

        >>> comtrade.pump.get_parameter_list("partnerAreas.json")

        Get the list of products

        >>> comtrade.pump.get_parameter_list("classificationHS.json")
        """
        url = self.url_metadata_base + file_name
        response = requests.get(url=url, headers=self.header, stream=True)
        print(f"HTTP response code: {response.status_code}")
        json_content = json.load(response.raw)
        df = pandas.json_normalize(json_content["results"])
        return df

    def update_db_parameter(self):
        """Update all parameter lists in the database table such as the product
        tables and country tables.

        Usage:

            >>> from biotrade.comtrade import comtrade
            >>> comtrade.pump.update_db_parameter()

        """
        # Dict for mapping table names and columns
        table_name_dict = {
            "cmd:HS": "product",
            "reporter": "reporter",
            "partner": "partner",
            "qtyunit": "quantity_unit",
            "flow": "flow",
            "mot": "modality_of_transport",
            "customs": "custom",
        }
        # First element of the ordered dict is the unique constraint column of the related table
        table_col_dict = {
            "cmd:HS": OrderedDict(
                {
                    "id": "product_code",
                    "text": "product_description",
                    "isLeaf": "is_leaf",
                    "aggrLevel": "aggregate_level",
                }
            ),
            "reporter": OrderedDict(
                {
                    "reporterCode": "reporter_code",
                    "reporterDesc": "reporter",
                    "reporterNote": "reporter_note",
                    "reporterCodeIsoAlpha2": "reporter_iso2",
                    "reporterCodeIsoAlpha3": "reporter_iso",
                    "entryEffectiveDate": "entry_effective_date",
                    "isGroup": "is_group",
                    "entryExpiredDate": "entry_expired_date",
                }
            ),
            "partner": OrderedDict(
                {
                    "PartnerCode": "partner_code",
                    "PartnerDesc": "partner",
                    "partnerNote": "partner_note",
                    "PartnerCodeIsoAlpha2": "partner_iso2",
                    "PartnerCodeIsoAlpha3": "partner_iso",
                    "entryEffectiveDate": "entry_effective_date",
                    "isGroup": "is_group",
                    "entryExpiredDate": "entry_expired_date",
                }
            ),
            "qtyunit": OrderedDict(
                {
                    "qtyCode": "unit_code",
                    "qtyAbbr": "unit_abbreviation",
                    "qtyDescription": "unit",
                }
            ),
            "flow": OrderedDict({"id": "flow_code", "text": "flow"}),
            "mot": OrderedDict(
                {
                    "id": "mode_of_transport_code",
                    "text": "mode_of_transport",
                }
            ),
            "customs": OrderedDict(
                {
                    "id": "custom_proc_code",
                    "text": "custom_proc",
                }
            ),
        }
        # Load data of complementary Comtrade tables
        for table in table_col_dict.keys():
            df = comtradeapicall.getReference(table)
            df = df.rename(columns=table_col_dict[table])
            if "flow" in df.columns:
                # Rename column content to snake case using a compiled regex
                regex_pat = re.compile(r"\W+")
                df["flow"] = (
                    df["flow"].str.replace(regex_pat, "_", regex=True).str.lower()
                )
            # Delete existing data in the database
            self.logger.info("Dropping existing %s table.", table_name_dict[table])
            # session.execute(delete_stmt)
            with self.db.engine.connect() as conn:
                conn.execute(self.db.tables[table_name_dict[table]].delete())
            # Remove potential duplicated id codes of the new data
            duplicated = df.duplicated(list(table_col_dict[table].items())[0][1])
            if any(duplicated):
                self.logger.info(
                    "Dropping the following duplicated rows from %s new data:\n %s",
                    table_name_dict[table],
                    df[duplicated],
                )
                df = df[~duplicated]
            # Keep only columns actually present in the database structure
            cols_in_db = self.db.tables[table_name_dict[table]].columns.keys()
            # Warn about missing columns
            missing_cols = set(df.columns) - set(cols_in_db)
            if missing_cols:
                msg = "The following columns are present in the Comtrade metadata "
                msg += f"{missing_cols} but they are not present in the database "
                msg += f"{table_name_dict[table]} table structure\n {cols_in_db} \n "
                msg += f"or in the {table_name_dict[table]} renaming dict\n"
                msg += f"{table_col_dict[table]}"
                warnings.warn(msg)
            # Store the data in the database
            self.db.append(
                df[cols_in_db], table_name_dict[table], drop_description=False
            )
            self.logger.info("The %s table has been updated.", table_name_dict[table])

    def write_log(
        self,
        timedate="",
        table="",
        max="",
        type="",
        freq="",
        px="",
        ps="",
        r="",
        p="",
        rg="",
        cc="",
        fmt="",
        head="",
        token="",
        status="",
        rows="",
    ):
        """
        Write API parameter values used to query the comtrade
        repository and status of stored data into the table
        "biotrade_data/comtrade/pump_comtrade_table.csv". If the table does not
        exist, it is created
        """
        # create headers of data frame
        df = pandas.DataFrame(
            {
                "date": [],
                "table": [],
                "max": [],
                "type": [],
                "freq": [],
                "px": [],
                "ps": [],
                "r": [],
                "p": [],
                "rg": [],
                "cc": [],
                "fmt": [],
                "head": [],
                "token": [],
                "status": [],
                "rows": [],
            }
        )
        # check if table exists
        if self.csv_log_path.exists() is False:
            # create table if not exist
            df.to_csv(self.csv_log_path, mode="a", index=False)
        # create new row of df containing API parameter values
        df.loc[len(df)] = [
            timedate.strftime("%Y-%m-%d %H:%M %z"),
            str(table),
            str(max),
            str(type),
            str(freq),
            str(px),
            str(ps),
            str(r),
            str(p),
            str(rg),
            str(cc),
            str(fmt),
            str(head),
            str(token),
            str(status),
            str(rows),
        ]
        # write row to table
        df.to_csv(self.csv_log_path, mode="a", index=False, header=False)

    def download_to_db_reporter_loop(
        self, table_name, start_year, end_year, product_code, frequency
    ):
        """Download data from the Comtrade API for all reporters for the given
        list of product codes and years specified.

        :param table_name, str defining the comtrade table to upload data
        ("monthly", "yearly", "yearly_hs2")
        :param start_year, int, the initial year to download data
        :param end_year, int, the final year to download data
        :param product_code, list, product codes
        :param frequency, str, defines the frequency of data ("M" monthly,
        "A" annual)

        Download data for all product codes, all reporters and all partners
        choosing frequency (annual or monthly) and table to store data
        Upon download failure, give a message, wait 10 seconds then try to
        download again.
        In case of new failure, double the wait time until the download works again
        Store the data in the specified database.

        For example download data for palm oil data of HS code 151190, between 2016 and 2021
        storing them into the monthly table of database

        >>> from biotrade.comtrade import comtrade
        >>> comtrade.pump.download_to_db_reporter_loop(
                table_name = 'monthly',
                start_year = 2016,
                end_year = 2021,
                product_code = '151190',
                frequency = 'M',
            )
        """
        records_downloaded = 0
        reporters = self.parent.country_groups.reporters
        # list of years
        years = [str(i) for i in range(start_year, end_year + 1)]
        for reporter_code in reporters.id:
            reporter_name = reporters.text[reporters.id == reporter_code].to_string(
                index=False
            )
            # https://comtrade.un.org/data/doc/api
            # "1 request every second (per IP address or authenticated user)."
            time.sleep(2)

            # reset counter for loop of products
            product_counter = 0
            # rest counter for loop of periods
            period_counter = 0
            # define the number of product codes for API depending on the frequency
            if frequency == "A":
                nr_product = 15
            elif frequency == "M":
                nr_product = 5
            # Select the first 5 products code at a time to avoid row limits to download
            product_block = product_code[
                product_counter * nr_product : (product_counter + 1) * nr_product
            ]
            # no more products to query
            while product_block != []:
                # differentiate selection of the period based on the frequency
                if frequency == "A":
                    # selection of 5 years
                    period = years[period_counter * 5 : (period_counter + 1) * 5]
                elif frequency == "M":
                    # selection of 1 year(12 months)
                    period = years[period_counter : period_counter + 1]
                # no more periods to query
                if period == []:
                    # continue with the next slot of products code
                    product_counter += 1
                    period_counter = 0
                    product_block = product_code[
                        product_counter
                        * nr_product : (product_counter + 1)
                        * nr_product
                    ]
                    continue
                else:
                    # conversion to a string
                    period = ",".join(period)

                download_successful = False
                # Reset additional sleep time used in case of error
                sleep_time = 10
                # Instantiate data frame
                df = pandas.DataFrame()
                # Try to download doubling sleep time until it succeeds.
                # Abort when sleep time increases to more than an hour.
                while not download_successful and sleep_time < 3600:
                    try:
                        df = self.download_df(
                            max=100000,
                            freq=frequency,
                            ps=period,
                            r=reporter_code,
                            cc=",".join(product_block),
                        )
                        download_successful = True
                    except Exception as error_http:
                        sleep_time *= 2
                        self.logger.info(
                            "Failed to download %s \n %s. \nWaiting %s seconds...",
                            reporter_name,
                            error_http,
                            sleep_time,
                        )
                        # Trace API parameters and error HTTP into table
                        self.write_log(
                            timedate=datetime.datetime.now(
                                pytz.timezone("Europe/Rome")
                            ),
                            table=table_name,
                            max=100000,
                            type="C",
                            freq=frequency,
                            px="HS",
                            ps=period,
                            r=reporter_code,
                            p="all",
                            rg="all",
                            cc=",".join(product_block),
                            fmt="json",
                            head="M",
                            token=self.token,
                            status=error_http,
                            rows="",
                        )
                        time.sleep(sleep_time)
                # Store in the database store the message if it fails
                try:
                    self.db.append(df, table_name)
                    # Define the status of stored data
                    if len(df) > 0:
                        status = "stored"
                    elif len(df) == 0:
                        status = "empty"
                except Exception as error_db:
                    self.logger.info(
                        "Failed to store %s in the database\n %s",
                        reporter_name,
                        error_db,
                    )
                    status = error_db
                # Keep track of the country name and length of the data in the log file
                records_downloaded += len(df)
                self.logger.info(
                    "Downloaded %s records for %s (code %s).\n"
                    + "%s records downloaded in total.",
                    len(df),
                    reporter_name,
                    reporter_code,
                    records_downloaded,
                )
                # Trace API parameters and db status into table
                self.write_log(
                    timedate=datetime.datetime.now(pytz.timezone("Europe/Rome")),
                    table=table_name,
                    max=100000,
                    type="C",
                    freq=frequency,
                    px="HS",
                    ps=period,
                    r=reporter_code,
                    p="all",
                    rg="all",
                    cc=",".join(product_block),
                    fmt="json",
                    head="M",
                    token=self.token,
                    status=status,
                    rows=len(df),
                )
                # new period
                period_counter += 1

    def download_to_db(self, table, *args, **kwargs):
        """Download Comtrade data and store it into the given database table"""
        # You might need to set the chunk size argument if there is an error.
        df = self.download(*args, **kwargs)
        # Remove the lengthy product description (to be stored in a dedicated table)
        df.drop(columns=["commodity"], inplace=True, errors="ignore")

    def loop_on_reporters(self, product_code, table_name):
        """Download data from the Comtrade API for all reporters for the given product code.

        Download the last 5 years for one product, all reporters and all partners
        Upon download failure, give a message, wait 10 seconds then try to download again.
        In case of new failure, double the wait time until the download works again
        Store the data in the PostGreSQL database.

        For example download data for wood products 44

        >>> from biotrade.comtrade import comtrade
        >>> comtrade.pump.loop_on_reporters("44", "yearly_hs2")
        """
        records_downloaded = 0
        reporters = self.parent.country_groups.reporters
        # Download the last 5 years for one product, one reporter and all its partners
        year = datetime.datetime.today().year
        # Convert each element of the list to a string
        years = [str(year - i) for i in range(1, 6)]
        years = ",".join(years)
        for reporter_code in reporters.id:
            reporter_name = reporters.text[reporters.id == reporter_code].to_string(
                index=False
            )
            download_successful = False
            # https://comtrade.un.org/data/doc/api
            # "1 request every second (per IP address or authenticated user)."
            time.sleep(2)
            # Reset additional sleep time used in case of error
            sleep_time = 10
            # Try to download doubling sleep time until it succeeds.
            # Abort when sleep time increases to more than half an hour.
            while not download_successful and sleep_time < 1800:
                try:
                    df = self.download(
                        max="9000",
                        type="C",
                        freq="A",
                        px="HS",
                        ps=years,
                        r=reporter_code,
                        p="all",
                        rg="all",
                        cc=product_code,
                        fmt="json",
                        head="M",
                    )
                    download_successful = True
                except Exception as error:
                    sleep_time *= 2
                    self.logger.info(
                        "Failed to download %s \n %s. \nWaiting %s seconds...",
                        reporter_name,
                        error,
                        sleep_time,
                    )
                    time.sleep(sleep_time)
            # Remove the lengthy product description (to be stored in a dedicated table)
            df.drop(columns=["commodity"], inplace=True, errors="ignore")
            # Store in the database store the message if it fails
            try:
                self.parent.db.append(df, table_name)
            # TODO check the error details,
            # check if data are the sames otherwise raise an error
            except Exception as error:
                self.logger.info(
                    "Failed to store %s in the database\n %s",
                    reporter_name,
                    error,
                )
            # Keep track of the country name and length of the data in the log file
            records_downloaded += len(df)
            self.logger.info(
                "Downloaded %s records for %s (code %s).\n"
                + "%s records downloaded in total.",
                len(df),
                reporter_name,
                reporter_code,
                records_downloaded,
            )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Define URL request headers
# Based on https://stackoverflow.com/a/66591873/2641825
HEADER = {
    "User-Agent": "Mozilla/5.0 (X11; Linux x86_64) "
    "AppleWebKit/537.11 (KHTML, like Gecko) "
    "Chrome/23.0.1271.64 Safari/537.11",
    "Accept": "text/html,application/xhtml+xml,application/xml;q=0.9,*/*;q=0.8",
    "Accept-Charset": "ISO-8859-1,utf-8;q=0.7,*;q=0.3",
    "Accept-Encoding": "none",
    "Accept-Language": "en-US,en;q=0.8",
    "Connection": "keep-alive",
}



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Written by Paul Rougieux and Selene Patani.

Copyright (c) 2023 European Union
Licenced under the MIT licence

JRC biomass Project.
Unit D1 Bioeconomy.
"""


class Coefficients:
    """
    Conversion coefficients for industrial processing and land area.
    """

    def __init__(self, parent):
        # Default attributes #
        self.parent = parent
        self.config_data_dir = self.parent.config_data_dir



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Written by Paul Rougieux and Selene Patani.

Copyright (c) 2023 European Union
Licenced under the MIT licence

JRC biomass Project.
Unit D1 Bioeconomy.


"""
# Built-in modules

# Third party modules
import logging
import pandas
from sqlalchemy import distinct, select
import warnings

# Internal modules


class Dump:
    """
    Dump comtrade data to compressed PostgreSQL dump files.


    For example dump forest related products (on a server)

        >>> from biotrade.comtrade import comtrade
        >>> comtrade.dump.store2d(table="monthly", 44)

    By default it will create a file name "monthly_44.dump.gz" in the biotrade_data/ repository
    The location of that repository is defined by the environment variable $BIOTRADE_DATA.

    Load the dump into a database (on a laptop)

        >>> file_path = comtrade.dump.data_dir / "raw_comtrade_monthly_15.csv.gz"
        >>> comtrade.dump.load_2d_csv("monthly", file_path)

    """

    # Log debug and error messages
    logger = logging.getLogger("biotrade.comtrade")

    def __init__(self, parent):
        # Default attributes #
        self.parent = parent
        self.db = self.parent.db
        # Location of the data
        self.data_dir = self.parent.data_dir / "dump"
        if not self.data_dir.exists():
            self.data_dir.mkdir()

    def store_2d_csv(
        self,
        table,
        product_code=None,
        chunk_size=10**6,
    ):
        """Dump rows where the product code starts with the given product code at the 2 digit level

        :param table_name, str defining the comtrade table
            ("monthly", "yearly", "yearly_hs2")
        :param int product_code 2 digit product code
        :param int chunk_size number of rows to read from the database at a time, to avoid memory errors

        If a list of product codes is given, store one file for each 2 digit code.

        For example save monthly or yearly "Animal Or Vegetable Fats And Oil"
        data which starts with 15:

            >>> from biotrade.comtrade import comtrade
            >>> comtrade.dump.store_2d_csv("monthly", 15)
            >>> comtrade.dump.store_2d_csv("yearly", 15)

        Dump all bioeconomy related products

            >>> hs2d = comtrade.products.hs2d.query("bioeconomy == 1")
            >>> for this_code in hs2d["product_code"]:
            >>>     comtrade.dump.store_2d_csv("monthly", this_code)

        """
        # If product_code is a single item, make it a list
        if isinstance(product_code, (int, str)):
            product_code = [product_code]
        # SQL Alchemy table object
        table = self.db.tables[table]
        # Loop on products
        for this_code in product_code:
            # Find products that start with the given code
            stmt = table.select()
            stmt = stmt.where(table.c.product_code.ilike(f"{this_code}%"))
            df_iter = pandas.read_sql_query(stmt, self.db.engine, chunksize=chunk_size)
            # Store the data frame to a dump file
            file_name = f"{table}_{this_code}.csv".replace(".", "_", 1)
            file_path = self.data_dir / (file_name + ".gz")
            # Remove the destination file if it exists
            if file_path.exists():
                file_path.unlink()
            self.logger.info("Dumping data in chunks")
            # Take a chunk of data from the database and write it to CSV, iterate
            for df in df_iter:
                self.logger.info(
                    "product %s, period %s ",
                    *df.loc[0, ["product_code", "period"]].to_list(),
                )
                df.to_csv(
                    file_path,
                    mode="a",
                    header=not file_path.exists(),
                    index=False,
                    compression="gzip",
                )
            self.logger.info("Stored data to:\n %s", file_path)

    def store_all_2d_csv(self, table):
        """Dump all bioeconomy related products

        Usage:

            >>> from biotrade.comtrade import comtrade
            >>> comtrade.dump.store_all_2d_csv("monthly")
            >>> comtrade.dump.store_all_2d_csv("yearly")
        """
        # Select 2 digit codes present in the database table
        stmt = select(distinct(self.db.tables[table].c.product_code))
        df = pandas.read_sql_query(stmt, self.db.engine)
        codes = [x for x in df.product_code.str[:2]]
        # Create one dump file for each 2 d code
        for this_code in set(codes):
            self.store_2d_csv(table, this_code)

    def load_2d_csv(self, table, file_path):
        """Load a dump into the given database table

        :param (str) table, name of the database table
        :param (str or file object) file_path, path to the file from which to save the data

        TODO: current version requires the user to manually delete the data
        from the database before an update

                psql $BIOTRADE_DATABASE_URL
                DELETE FROM raw_comtrade.monthly WHERE product_code like '44%';
                DELETE FROM raw_comtrade.yearly WHERE product_code like '44%';

        If data is already present in the database for that product code, the
        user will be asked to confirm deletion. To avoid this confirmation
        message when processing many dump files automatically, delete all table
        content before calling this function.

        For information, in case the data is not deleted and if the new data causes
        duplications the database unique constraint will raise an error:

            IntegrityError: (psycopg2.errors.UniqueViolation) duplicate key value
            violates unique constraint
            "monthly_period_flow_code_reporter_code_partner_code_product_key"
            DETAIL:  Key (period, flow_code, reporter_code, partner_code,
            product_code, unit_code, flag)=(201605, 1, 372, 360, 442090, 0, 0)

        Usage:

            >>> from biotrade.comtrade import comtrade
            >>> # Load monthly data for one 2 digit level code and all child codes under it
            >>> file_path = comtrade.dump.data_dir / "raw_comtrade_monthly_15.csv.gz"
            >>> comtrade.dump.load_2d_csv("monthly", file_path)
            >>> # Load yearly data for many 2 digit level codes and all child codes under it
            >>> periodicity = "yearly"
            >>> for code in ["12", "15", "23"]:
            >>>     file_path = comtrade.dump.data_dir
            >>>     file_path = file_path / f"raw_comtrade_{periodicity}_{code}.csv.gz"
            >>>     comtrade.dump.load_2d_csv(periodicity, file_path)

        """
        # Read the csv file in memory
        df = pandas.read_csv(
            file_path,
            # Force the id column to remain a character column,
            # otherwise str "01" becomes int 1.
            dtype={"product_code": str},
        )
        # Temporary fix because of mismatch between the latest Comtrade data
        # and the database table structure
        db_cols = self.db.tables["yearly"].c.keys()
        msg = "Columns present in the dump file but not in the database:"
        msg += f"{set(df.columns) - set(db_cols)}"
        warnings.warn(msg)
        for col in df.columns:
            if col not in db_cols:
                df.drop(columns=col, inplace=True)
        # Check that the name of the file contains the unique product code in the data

        # Delete existing data for the corresponding product code
        # TODO: uptate the db.delete method so that it can deal with product_code.
        # This should be possible if the where statements can be concatenated
        # Or create the statement here to delete at the HS 2 to level.
        self.logger.info("Reading into a data frame:\n %s", file_path)

        # Write to the database
        self.db.append(df, table)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Comparison between FAOSTAT and Comtrade country codes

    >>> from biotrade.comtrade import comtrade
    >>> from biotrade.faostat import faostat
    >>> cgc = comtrade.country_groups.reporters
    >>> cgf = faostat.country_groups.df[["faost_code", "un_code", "fao_table_name"]]
    >>> compare = cgc.merge(cgf, left_on="id", right_on="un_code")
    >>> # Show only countries where names don't match
    >>> compare.query("text != fao_table_name")

TODO: Add a df method and inherit from common/country_groups, see issue #102

"""
# Third party modules
import pandas


class CountryGroups(object):
    """
    Comtrade country lists.
    """

    def __init__(self, parent):
        # Default attributes #
        self.parent = parent
        # Directories #
        self.config_data_dir = self.parent.config_data_dir

    @property
    def reporters(self):
        """The module internal list of reporter countries
        Usage:

        >>> from biotrade.comtrade import comtrade
        >>> comtrade.country_groups.reporters

        For information the internal list is a filtered version of
        the list of reporters
        originally downloaded from Comtrade with the method:

        >>> comtrade.pump.get_parameter_list("reporterAreas.json")
        """
        df = pandas.read_csv(self.config_data_dir / "comtrade_reporters.csv")
        # Remove the special id "all"
        df = df[df["id"] != "all"]
        # Change id to a numerical variable
        df["id"] = pandas.to_numeric(df["id"], errors="coerce")
        return df
